@@ -1,7 +1,7 @@
 /*******************************************************************************
    This file is part of Distributed Privacy Guard (DKGPG).
 
- Copyright (C) 2018  Heiko Stamer <HeikoStamer@gmx.net>
+ Copyright (C) 2018, 2019  Heiko Stamer <HeikoStamer@gmx.net>
 
    DKGPG is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -160,17 +160,23 @@ void run_instance
 		delete ring;
 		exit(-1);
 	}
+	delete ring;
 	if (!prv->pub->valid || ((opt_y == NULL) && prv->Weak(opt_verbose)))
 	{
 		std::cerr << "ERROR: primary key is invalid or weak" << std::endl;
-		delete ring;
 		delete prv;
 		exit(-1);
 	}
 	if ((prv->pkalgo != TMCG_OPENPGP_PKALGO_EXPERIMENTAL7) && (opt_y == NULL))
 	{
 		std::cerr << "ERROR: primary key is not a tDSS/DSA key" << std::endl;
-		delete ring;
+		delete prv;
+		exit(-1);
+	}
+	if ((prv->pub->AccumulateFlags() & 0x0800) != 0x0800)
+	{
+		std::cerr << "ERROR: primary key is not intented for timestamping" <<
+			std::endl;
 		delete prv;
 		exit(-1);
 	}
@@ -181,7 +187,6 @@ void run_instance
 	{
 		if (!read_message(opt_i, armored_signature))
 		{
-			delete ring;
 			delete prv;
 			exit(-1);
 		}
@@ -219,7 +224,6 @@ void run_instance
 			std::cerr << "ERROR: wrong signature type " <<
 				(int)signature->type << " found" << std::endl;
 			delete signature;
-			delete ring;
 			delete prv;
 			exit(-1);
 		}
@@ -228,7 +232,6 @@ void run_instance
 		{
 			std::cerr << "ERROR: cannot extract signature body" << std::endl;
 			delete signature;
-			delete ring;
 			delete prv;
 			exit(-1);
 		}
@@ -237,7 +240,6 @@ void run_instance
 	{
 		std::cerr << "ERROR: cannot parse resp. use the provided signature" <<
 			std::endl;
-		delete ring;
 		delete prv;
 		exit(-1);
 	}
@@ -258,7 +260,6 @@ void run_instance
 		{
 			delete dss;
 			delete signature;
-			delete ring;
 			delete prv;
 			exit(-1);
 		}
@@ -269,7 +270,6 @@ void run_instance
 				std::endl;
 			delete dss;
 			delete signature;
-			delete ring;
 			delete prv;
 			exit(-1);
 		}
@@ -284,12 +284,11 @@ void run_instance
 				std::string pwd;
 				if (!TMCG_ParseHelper::gs(passwords, '/', pwd))
 				{
-					std::cerr << "ERROR: S_" << whoami << ": " <<
+					std::cerr << "ERROR: p_" << whoami << ": " <<
 						"cannot read" << " password for protecting channel" <<
-						" to S_" << i << std::endl;
+						" to p_" << i << std::endl;
 					delete dss;
 					delete signature;
-					delete ring;
 					delete prv;
 					exit(-1);
 				}
@@ -297,12 +296,11 @@ void run_instance
 				if (((i + 1) < peers.size()) &&
 					!TMCG_ParseHelper::nx(passwords, '/'))
 				{
-					std::cerr << "ERROR: S_" << whoami << ": " << "cannot" <<
-						" skip to next password for protecting channel to S_" <<
+					std::cerr << "ERROR: p_" << whoami << ": " << "cannot" <<
+						" skip to next password for protecting channel to p_" <<
 						(i + 1) << std::endl;
 					delete dss;
 					delete signature;
-					delete ring;
 					delete prv;
 					exit(-1);
 				}
@@ -310,7 +308,7 @@ void run_instance
 			else
 			{
 				// simple key -- we assume that GNUnet provides secure channels
-				key << "dkg-timestamp::S_" << (i + whoami);
+				key << "dkg-timestamp::p_" << (i + whoami);
 			}
 			uP_in.push_back(pipefd[i][whoami][0]);
 			uP_out.push_back(pipefd[whoami][i][1]);
@@ -337,23 +335,7 @@ void run_instance
 				aiou2, aiounicast::aio_scheduler_roundrobin, (opt_W * 60));
 		rbc->setID(myID);
 		// perform a simple exchange test with debug output
-		for (size_t i = 0; i < num_xtests; i++)
-		{
-			mpz_t xtest;
-			mpz_init_set_ui(xtest, i);
-			std::cerr << "INFO: S_" << whoami << ": xtest = " << xtest <<
-				" <-> ";
-			rbc->Broadcast(xtest);
-			for (size_t ii = 0; ii < peers.size(); ii++)
-			{
-				if (!rbc->DeliverFrom(xtest, ii))
-					std::cerr << "<X> ";
-				else
-					std::cerr << xtest << " ";
-			}
-			std::cerr << std::endl;
-			mpz_clear(xtest);
-		}
+		xtest(num_xtests, whoami, peers.size(), rbc);
 		// participants must agree on a common signature creation time (OpenPGP)
 		if (opt_verbose)
 			std::cerr << "INFO: agree on a signature creation time for" <<
@@ -375,8 +357,8 @@ void run_instance
 				}
 				else
 				{
-					std::cerr << "WARNING: S_" << whoami << ": no signature" <<
-						" creation timestamp received from S_" << i << std::endl;
+					std::cerr << "WARNING: p_" << whoami << ": no signature" <<
+						" creation timestamp received from p_" << i << std::endl;
 				}
 			}
 		}
@@ -384,19 +366,18 @@ void run_instance
 		std::sort(tvs.begin(), tvs.end());
 		if (tvs.size() < (peers.size() - T_RBC))
 		{
-			std::cerr << "ERROR: S_" << whoami << ": not enough timestamps" <<
+			std::cerr << "ERROR: p_" << whoami << ": not enough timestamps" <<
 				" received" << std::endl;
 			delete rbc, delete aiou, delete aiou2;
 			delete dss;
 			delete signature;
-			delete ring;
 			delete prv;
 			exit(-1);
 		}
 		// use a median value as some kind of gentle agreement
 		csigtime = tvs[tvs.size()/2];
 		if (opt_verbose)
-			std::cerr << "INFO: S_" << whoami << ": canonicalized signature" <<
+			std::cerr << "INFO: p_" << whoami << ": canonicalized signature" <<
 				" creation time (timestamp) = " << csigtime << std::endl;
 		if (opt_a)
 		{
@@ -406,25 +387,18 @@ void run_instance
 			else
 				acc = (hst - csigtime); // set timestamp accuracy
 			if (opt_verbose)
-				std::cerr << "INFO: S_" << whoami << ": set accuracy = " <<
+				std::cerr << "INFO: p_" << whoami << ": set accuracy = " <<
 					(unsigned long int)acc << std::endl;
 		}
 		// select hash algorithm for OpenPGP based on |q| (size in bit)
-		if (mpz_sizeinbase(dss->q, 2L) == 256)
-			hashalgo = TMCG_OPENPGP_HASHALGO_SHA256; // SHA256 (alg 8)
-		else if (mpz_sizeinbase(dss->q, 2L) == 384)
-			hashalgo = TMCG_OPENPGP_HASHALGO_SHA384; // SHA384 (alg 9)
-		else if (mpz_sizeinbase(dss->q, 2L) == 512)
-			hashalgo = TMCG_OPENPGP_HASHALGO_SHA512; // SHA512 (alg 10)
-		else
+		if (!select_hashalgo(dss, hashalgo))
 		{
-			std::cerr << "ERROR: S_" << whoami << ": selecting hash" <<
+			std::cerr << "ERROR: p_" << whoami << ": selecting hash" <<
 				" algorithm failed for |q| = " << mpz_sizeinbase(dss->q, 2L) <<
 				std::endl;
 			delete rbc, delete aiou, delete aiou2;
 			delete dss;
 			delete signature;
-			delete ring;
 			delete prv;
 			exit(-1);
 		}
@@ -435,9 +409,9 @@ void run_instance
 		hashalgo = TMCG_OPENPGP_HASHALGO_SHA512;
 	}
 
-	// compute the trailer and the hash from the signature
+	// compute the trailer and the hash of the signature
 	if (opt_verbose)
-		std::cerr << "INFO: constructing timestamp signature" << std::endl;
+		std::cerr << "INFO: constructing the timestamp signature" << std::endl;
 	tmcg_openpgp_octets_t trailer, hash, left, issuer;
 	CallasDonnerhackeFinneyShawThayerRFC4880::
 		FingerprintCompute(prv->pub->pub_hashing, issuer);
@@ -446,7 +420,10 @@ void run_instance
 	if (opt_a)
 	{
 		if (opt_verbose)
-			std::cerr << "INFO: include an OpenPGP notation" << std::endl;
+		{
+			std::cerr << "INFO: include an OpenPGP notation on accuracy" <<
+				std::endl;
+		}
 		std::string accuracy_name = "accuracy@dkg-timestamp";
 		std::stringstream avs;
 		avs << (unsigned long int)acc;
@@ -459,11 +436,14 @@ void run_instance
 	}
 	if (sn.length() != 0)
 	{
-		if (opt_verbose)
-			std::cerr << "INFO: include an OpenPGP notation" << std::endl;
 		size_t dpos = sn.find(":");
 		if ((dpos != sn.npos) && (dpos > 0) && ((sn.length() - dpos) > 1))
 		{
+			if (opt_verbose)
+			{
+				std::cerr << "INFO: include an OpenPGP notation on S/N" <<
+					std::endl;
+			}
 			std::string serialnumber_name = sn.substr(0, dpos);
 			std::string serialnumber_value = sn.substr(dpos + 1,
 				sn.length() - dpos - 1);
@@ -474,8 +454,10 @@ void run_instance
 			notations.push_back(serialnumber);
 		}
 		else
+		{
 			std::cerr << "WARNING: wrong delimiter position for given" <<
-				" OpenPGP notation" << std::endl;
+				" OpenPGP notation; ignored" << std::endl;
+		}
 	} // TODO: option -t --target => use other variant of TimestampSignature
 	  //       with hash value supplied by caller, cf. [RFC 3161]
 	if (opt_y == NULL)
@@ -495,180 +477,28 @@ void run_instance
 	CallasDonnerhackeFinneyShawThayerRFC4880::
 		StandaloneHash(trailer, hashalgo, hash, left);
 
-	// prepare the hash value
-	tmcg_openpgp_byte_t buffer[1024];
-	size_t buflen = 0;
-	memset(buffer, 0, sizeof(buffer));
-	if (opt_verbose > 1)
-		std::cerr << std::hex << "INFO: hash = ";
-	for (size_t i = 0; i < hash.size(); i++, buflen++)
-	{
-		if (i < sizeof(buffer))
-			buffer[i] = hash[i];
-		if (opt_verbose > 1)
-			std::cerr << (int)hash[i] << " ";
-	}
-	if (opt_verbose > 1)
-		std::cerr << std::dec << std::endl;
-
 	// sign the hash value
-	gcry_error_t ret;
-	gcry_mpi_t r, s;
-	r = gcry_mpi_new(2048);
-	s = gcry_mpi_new(2048);
-	if (opt_y == NULL)
-	{
-		gcry_mpi_t h;
-		ret = gcry_mpi_scan(&h, GCRYMPI_FMT_USG, buffer, buflen, NULL);
-		if (ret)
-		{
-			std::cerr << "ERROR: S_" << whoami << ": gcry_mpi_scan() failed" <<
-				" for h" << std::endl;
-			delete rbc, delete aiou, delete aiou2;
-			delete dss;
-			delete signature;
-			delete ring;
-			delete prv;
-			exit(-1);
-		}
-		if (opt_verbose > 1)
-			std::cerr << "INFO: S_" << whoami << ": h = " << h << std::endl;
-		mpz_t dsa_m, dsa_r, dsa_s;
-		mpz_init(dsa_m), mpz_init(dsa_r), mpz_init(dsa_s);
-		if (!tmcg_mpz_set_gcry_mpi(h, dsa_m))
-		{
-			std::cerr << "ERROR: S_" << whoami << ": tmcg_mpz_set_gcry_mpi()" <<
-				" failed for dsa_m" << std::endl;
-			gcry_mpi_release(r), gcry_mpi_release(s), gcry_mpi_release(h);
-			mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-			delete rbc, delete aiou, delete aiou2;
-			delete dss;
-			delete signature;
-			delete ring;
-			delete prv;
-			exit(-1);
-		}
-		gcry_mpi_release(h);
-		std::stringstream err_log_sign;
-		if (opt_verbose)
-			std::cerr << "INFO: S_" << whoami << ": dss.Sign()" << std::endl;
-		if (!dss->Sign(peers.size(), whoami, dsa_m, dsa_r, dsa_s,
-			prv->tdss_idx2dkg, prv->tdss_dkg2idx, aiou, rbc, err_log_sign))
-		{
-			std::cerr << "ERROR: S_" << whoami << ": " <<
-				"tDSS Sign() failed" << std::endl;
-			std::cerr << "ERROR: S_" << whoami << ": log follows " <<
-				std::endl << err_log_sign.str();
-			gcry_mpi_release(r), gcry_mpi_release(s);
-			mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-			delete rbc, delete aiou, delete aiou2;
-			delete dss;
-			delete signature;
-			delete ring;
-			delete prv;
-			exit(-1);
-		}
-		if (opt_verbose > 1)
-			std::cerr << "INFO: S_" << whoami << ": log follows " <<
-				std::endl << err_log_sign.str();
-		if (!tmcg_mpz_get_gcry_mpi(r, dsa_r))
-		{
-			std::cerr << "ERROR: S_" << whoami << ": tmcg_mpz_get_gcry_mpi()" <<
-				" failed for dsa_r" << std::endl;
-			gcry_mpi_release(r), gcry_mpi_release(s);
-			mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-			delete rbc, delete aiou, delete aiou2;
-			delete dss;
-			delete signature;
-			delete ring;
-			delete prv;
-			exit(-1);
-		}
-		if (!tmcg_mpz_get_gcry_mpi(s, dsa_s))
-		{
-			std::cerr << "ERROR: S_" << whoami << ": tmcg_mpz_get_gcry_mpi()" <<
-				" failed for dsa_s" << std::endl;
-			gcry_mpi_release(r), gcry_mpi_release(s);
-			mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-			delete rbc, delete aiou, delete aiou2;
-			delete dss;
-			delete signature;
-			delete ring;
-			delete prv;
-			exit(-1);
-		}
-		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-	}
-	else
-	{
-		switch (prv->pkalgo)
-		{
-			case TMCG_OPENPGP_PKALGO_RSA:
-			case TMCG_OPENPGP_PKALGO_RSA_SIGN_ONLY:
-				ret = CallasDonnerhackeFinneyShawThayerRFC4880::
-					AsymmetricSignRSA(hash, prv->private_key, hashalgo, s);
-				break;
-			case TMCG_OPENPGP_PKALGO_DSA:
-				ret = CallasDonnerhackeFinneyShawThayerRFC4880::
-					AsymmetricSignDSA(hash, prv->private_key, r, s);
-				break;
-			case TMCG_OPENPGP_PKALGO_ECDSA:
-				ret = CallasDonnerhackeFinneyShawThayerRFC4880::
-					AsymmetricSignECDSA(hash, prv->private_key, r, s);
-				break;
-			default:
-				std::cerr << "ERROR: public-key algorithm " <<
-					(int)prv->pkalgo << " not supported" << std::endl;
-				gcry_mpi_release(r), gcry_mpi_release(s);
-				delete signature;
-				delete ring;
-				delete prv;
-				exit(-1);
-		}
-		if (ret)
-		{
-			std::cerr << "ERROR: signing of hash value failed " <<
-				"(rc = " << gcry_err_code(ret) << ", str = " <<
-				gcry_strerror(ret) << ")" << std::endl;
-			gcry_mpi_release(r), gcry_mpi_release(s);
-			delete signature;
-			delete ring;
-			delete prv;
-			exit(-1);
-		}
-	}
 	tmcg_openpgp_octets_t sig;
-	switch (prv->pkalgo)
+	if (!sign_hash(hash, trailer, left, whoami, peers.size(), prv, hashalgo,
+		sig, opt_verbose, opt_y, dss, aiou, rbc))
 	{
-		case TMCG_OPENPGP_PKALGO_RSA:
-		case TMCG_OPENPGP_PKALGO_RSA_SIGN_ONLY:
-			CallasDonnerhackeFinneyShawThayerRFC4880::
-				PacketSigEncode(trailer, left, s, sig);
-			break;
-		case TMCG_OPENPGP_PKALGO_DSA:
-		case TMCG_OPENPGP_PKALGO_ECDSA:
-		case TMCG_OPENPGP_PKALGO_EXPERIMENTAL7:
-			CallasDonnerhackeFinneyShawThayerRFC4880::
-				PacketSigEncode(trailer, left, r, s, sig);
-			break;
-		default:
-			std::cerr << "ERROR: public-key algorithm " << (int)prv->pkalgo <<
-				" not supported" << std::endl;
-			gcry_mpi_release(r), gcry_mpi_release(s);
-			delete signature;
-			delete ring;
-			delete prv;
-			exit(-1);
+		if (opt_y == NULL)
+		{
+			delete rbc, delete aiou, delete aiou2;
+			delete dss;
+		}
+		delete signature;
+		delete prv;
+		exit(-1);
 	}
-	gcry_mpi_release(r), gcry_mpi_release(s);
 
 	// release allocated ressources
 	if (opt_y == NULL)
 	{
 		// at the end: deliver some more rounds for still waiting parties
-		time_t synctime = aiounicast::aio_timeout_long;
+		time_t synctime = (opt_W * 6);
 		if (opt_verbose)
-			std::cerr << "INFO: S_" << whoami << ": waiting approximately " <<
+			std::cerr << "INFO: p_" << whoami << ": waiting approximately " <<
 				(synctime * (T_RBC + 1)) << " seconds for stalled parties" <<
 				std::endl;
 		rbc->Sync(synctime);
@@ -677,14 +507,14 @@ void run_instance
 		// release handles (unicast channel)
 		if (opt_verbose)
 		{
-			std::cerr << "INFO: S_" << whoami << ": unicast channels";
+			std::cerr << "INFO: p_" << whoami << ": unicast channels";
 			aiou->PrintStatistics(std::cerr);
 			std::cerr << std::endl;
 		}
 		// release handles (broadcast channel)
 		if (opt_verbose)
 		{
-			std::cerr << "INFO: S_" << whoami << ": broadcast channel";
+			std::cerr << "INFO: p_" << whoami << ": broadcast channel";
 			aiou2->PrintStatistics(std::cerr);
 			std::cerr << std::endl;
 		}
@@ -694,7 +524,6 @@ void run_instance
 		delete dss;
 	}
 	delete signature;
-	delete ring;
 	delete prv;
 
 	// output the result
@@ -736,7 +565,7 @@ void fork_instance
 	{
 		if (pid[whoami] == 0)
 		{
-			/* BEGIN child code: participant S_i */
+			/* BEGIN child code: participant p_i */
 			time_t sigtime = time(NULL);
 #ifdef GNUNET
 			run_instance(whoami, sigtime, gnunet_opt_xtests);
@@ -744,9 +573,9 @@ void fork_instance
 			run_instance(whoami, sigtime, 0);
 #endif
 			if (opt_verbose)
-				std::cerr << "INFO: S_" << whoami << ": exit(0)" << std::endl;
+				std::cerr << "INFO: p_" << whoami << ": exit(0)" << std::endl;
 			exit(0);
-			/* END child code: participant S_i */
+			/* END child code: participant p_i */
 		}
 		else
 		{
@@ -831,19 +660,19 @@ int main
 		),
 		GNUNET_GETOPT_option_uint('w',
 			"wait",
-			"TIME",
+			"INTEGER",
 			"minutes to wait until start of signing protocol",
 			&gnunet_opt_wait
 		),
 		GNUNET_GETOPT_option_uint('W',
 			"aiou-timeout",
-			"TIME",
+			"INTEGER",
 			"timeout for point-to-point messages in minutes",
 			&gnunet_opt_W
 		),
 		GNUNET_GETOPT_option_string('y',
 			"yaot",
-			"FILNAME",
+			"FILENAME",
 			"yet another OpenPGP tool with private key in FILENAME",
 			&gnunet_opt_y
 		),
@@ -1001,7 +830,7 @@ int main
 					std::endl;
 				std::cout << "  -V, --verbose  turn on verbose output" <<
 					std::endl;
-				std::cout << "  -W TIME        timeout for point-to-point" <<
+				std::cout << "  -W INTEGER     timeout for point-to-point" <<
 					" messages in minutes" << std::endl;
 				std::cout << "  -y FILENAME    yet another OpenPGP tool with" <<
 					" private key in FILENAME" << std::endl;
@@ -1040,9 +869,9 @@ int main
 		}
 	}
 #ifdef DKGPG_TESTSUITE
-	peers.push_back("Test2");
-	peers.push_back("Test3");
-	peers.push_back("Test4");
+	peers.push_back("TestTS2");
+	peers.push_back("TestTS3");
+	peers.push_back("TestTS4");
 	ifilename = "Test1_output.sig";
 	opt_i = (char*)ifilename.c_str();
 	ofilename = "Test1_output_timestamp.sig";
@@ -1078,12 +907,7 @@ int main
 			usage << std::endl;
 		return -1;
 	}
-
-	// canonicalize peer list
-	std::sort(peers.begin(), peers.end());
-	std::vector<std::string>::iterator it =
-		std::unique(peers.begin(), peers.end());
-	peers.resize(std::distance(peers.begin(), it));
+	canonicalize(peers);
 	if (((peers.size() < 3)  || (peers.size() > DKGPG_MAX_N)) && (opt_y == NULL))
 	{
 		std::cerr << "ERROR: too few or too many peers given" << std::endl;
@@ -1097,7 +921,7 @@ int main
 	}
 
 	// lock memory
-	bool force_secmem = false;
+	bool force_secmem = false, should_unlock = false;
 	if (!lock_memory())
 	{
 		std::cerr << "WARNING: locking memory failed; CAP_IPC_LOCK required" <<
@@ -1105,209 +929,134 @@ int main
 		// at least try to use libgcrypt's secure memory
 		force_secmem = true;
 	}
+	else
+		should_unlock = true;
 
 	// initialize LibTMCG
 	if (!init_libTMCG(force_secmem))
 	{
 		std::cerr << "ERROR: initialization of LibTMCG failed" << std::endl;
+		if (should_unlock)
+			unlock_memory();
 		return -1;
 	}
 	if (opt_verbose)
 		std::cerr << "INFO: using LibTMCG version " << version_libTMCG() <<
 			std::endl;
 	
-	// initialize return code
+	// initialize return code and do the main work
 	int ret = 0;
-	// create underlying point-to-point channels, if built-in TCP/IP requested
 	if ((opt_hostname != NULL) && (opt_y == NULL))
 	{
-		if (port.length())
-			opt_p = strtoul(port.c_str(), NULL, 10); // start port from options
-		if ((opt_p < 1) || (opt_p > 65535))
-		{
-			std::cerr << "ERROR: no valid TCP start port given" << std::endl;
-			return -1;
-		}
-		tcpip_init(hostname);
-		tcpip_bindports((uint16_t)opt_p, false);
-		tcpip_bindports((uint16_t)opt_p, true);
-		while (tcpip_connect((uint16_t)opt_p, false) < peers.size())
-			sleep(1);
-		while (tcpip_connect((uint16_t)opt_p, true) < peers.size())
-			sleep(1);
-		tcpip_accept();
-		tcpip_fork();
-		ret = tcpip_io();
-		tcpip_close();
-		tcpip_done();
-		return ret;
+		// start interactive variant, if built-in TCP/IP requested
+		ret = run_tcpip(peers.size(), opt_p, hostname, port);
 	}
 	else if (opt_y != NULL)
 	{
-		// run as replacement for GnuPG et al. (yet-another-openpgp-tool)
-		run_instance(0, time(NULL), 0);
-		return ret;
+		// start a single instance as replacement for GnuPG et al.
+		fork_instance(0);
+		ret = wait_instance(0, opt_verbose, pid);
 	}
-
-	// start interactive variant with GNUnet or otherwise a local test
-#ifdef GNUNET
-	static const struct GNUNET_GETOPT_CommandLineOption myoptions[] = {
-		GNUNET_GETOPT_option_flag('a',
-			"accuracy",
-			"include OpenPGP notation that represents time deviation",
-			&gnunet_opt_a
-		),
-		GNUNET_GETOPT_option_string('H',
-			"hostname",
-			"STRING",
-			"hostname (e.g. onion address) of this peer within PEERS",
-			&gnunet_opt_hostname
-		),
-		GNUNET_GETOPT_option_string('i',
-			"input",
-			"FILENAME",
-			"read target signature from FILENAME",
-			&gnunet_opt_i
-		),
-		GNUNET_GETOPT_option_string('k',
-			"keyring",
-			"FILENAME",
-			"use keyring FILENAME containing external revocation keys",
-			&gnunet_opt_k
-		),
-		GNUNET_GETOPT_option_string('o',
-			"output",
-			"FILENAME",
-			"write generated timestamp signature to FILENAME",
-			&gnunet_opt_o
-		),
-		GNUNET_GETOPT_option_string('p',
-			"port",
-			"STRING",
-			"GNUnet CADET port to listen/connect",
-			&gnunet_opt_port
-		),
-		GNUNET_GETOPT_option_string('P',
-			"passwords",
-			"STRING",
-			"exchanged passwords to protect private and broadcast channels",
-			&gnunet_opt_passwords
-		),
-		GNUNET_GETOPT_option_string('s',
-			"sn",
-			"KEY:VALUE",
-			"embed this OpenPGP notation (e.g. serial number)",
-			&gnunet_opt_s
-		),
-		GNUNET_GETOPT_option_string('U',
-			"URI",
-			"STRING",
-			"policy URI tied to timestamp signature",
-			&gnunet_opt_URI
-		),
-		GNUNET_GETOPT_option_flag('V',
-			"verbose",
-			"turn on verbose output",
-			&gnunet_opt_verbose
-		),
-		GNUNET_GETOPT_option_uint('w',
-			"wait",
-			"TIME",
-			"minutes to wait until start of signing protocol",
-			&gnunet_opt_wait
-		),
-		GNUNET_GETOPT_option_uint('W',
-			"aiou-timeout",
-			"TIME",
-			"timeout for point-to-point messages in minutes",
-			&gnunet_opt_W
-		),
-		GNUNET_GETOPT_option_string('y',
-			"yaot",
-			"FILNAME",
-			"yet another OpenPGP tool with private key in FILENAME",
-			&gnunet_opt_y
-		),
-		GNUNET_GETOPT_option_uint('x',
-			"x-tests",
-			NULL,
-			"number of exchange tests",
-			&gnunet_opt_xtests
-		),
-		GNUNET_GETOPT_OPTION_END
-	};
-	ret = GNUNET_PROGRAM_run(argc, argv, usage, about, myoptions, &gnunet_run,
-		argv[0]);
-//	GNUNET_free((void *) argv);
-	if (ret == GNUNET_OK)
-		return 0;
 	else
-		return -1;
+	{
+		// start interactive variant with GNUnet or otherwise a local test
+#ifdef GNUNET
+		static const struct GNUNET_GETOPT_CommandLineOption myoptions[] = {
+			GNUNET_GETOPT_option_flag('a',
+				"accuracy",
+				"include OpenPGP notation that represents time deviation",
+				&gnunet_opt_a
+			),
+			GNUNET_GETOPT_option_string('H',
+				"hostname",
+				"STRING",
+				"hostname (e.g. onion address) of this peer within PEERS",
+				&gnunet_opt_hostname
+			),
+			GNUNET_GETOPT_option_string('i',
+				"input",
+				"FILENAME",
+				"read target signature from FILENAME",
+				&gnunet_opt_i
+			),
+			GNUNET_GETOPT_option_string('k',
+				"keyring",
+				"FILENAME",
+				"use keyring FILENAME containing external revocation keys",
+				&gnunet_opt_k
+			),
+			GNUNET_GETOPT_option_string('o',
+				"output",
+				"FILENAME",
+				"write generated timestamp signature to FILENAME",
+				&gnunet_opt_o
+			),
+			GNUNET_GETOPT_option_string('p',
+				"port",
+				"STRING",
+				"GNUnet CADET port to listen/connect",
+				&gnunet_opt_port
+			),
+			GNUNET_GETOPT_option_string('P',
+				"passwords",
+				"STRING",
+				"exchanged passwords to protect private and broadcast channels",
+				&gnunet_opt_passwords
+			),
+			GNUNET_GETOPT_option_string('s',
+				"sn",
+				"KEY:VALUE",
+				"embed this OpenPGP notation (e.g. serial number)",
+				&gnunet_opt_s
+			),
+			GNUNET_GETOPT_option_string('U',
+				"URI",
+				"STRING",
+				"policy URI tied to timestamp signature",
+				&gnunet_opt_URI
+			),
+			GNUNET_GETOPT_option_flag('V',
+				"verbose",
+				"turn on verbose output",
+				&gnunet_opt_verbose
+			),
+			GNUNET_GETOPT_option_uint('w',
+				"wait",
+				"INTEGER",
+				"minutes to wait until start of signing protocol",
+				&gnunet_opt_wait
+			),
+			GNUNET_GETOPT_option_uint('W',
+				"aiou-timeout",
+				"INTEGER",
+				"timeout for point-to-point messages in minutes",
+				&gnunet_opt_W
+			),
+			GNUNET_GETOPT_option_string('y',
+				"yaot",
+				"FILENAME",
+				"yet another OpenPGP tool with private key in FILENAME",
+				&gnunet_opt_y
+			),
+			GNUNET_GETOPT_option_uint('x',
+				"x-tests",
+				NULL,
+				"number of exchange tests",
+				&gnunet_opt_xtests
+			),
+			GNUNET_GETOPT_OPTION_END
+		};
+		ret = GNUNET_PROGRAM_run(argc, argv, usage, about, myoptions,
+			&gnunet_run, argv[0]);
+		if (ret != GNUNET_OK)
+			ret = -1;
 #else
-	std::cerr << "WARNING: GNUnet CADET is required for the message exchange" <<
-		" of this program" << std::endl;
+		ret = run_localtest(peers.size(), opt_verbose, pid, pipefd,
+			broadcast_pipefd, &fork_instance);
 #endif
-
-	std::cerr << "INFO: running local test with " << peers.size() <<
-		" participants" << std::endl;
-	// open pipes
-	for (size_t i = 0; i < peers.size(); i++)
-	{
-		for (size_t j = 0; j < peers.size(); j++)
-		{
-			if (pipe(pipefd[i][j]) < 0)
-				perror("ERROR: dkg-timestamp (pipe)");
-			if (pipe(broadcast_pipefd[i][j]) < 0)
-				perror("ERROR: dkg-timestamp (pipe)");
-		}
 	}
-	
-	// start childs
-	for (size_t i = 0; i < peers.size(); i++)
-		fork_instance(i);
-
-	// sleep for five seconds
-	sleep(5);
-	
-	// wait for childs and close pipes
-	for (size_t i = 0; i < peers.size(); i++)
-	{
-		int wstatus = 0;
-		if (opt_verbose)
-			std::cerr << "INFO: waitpid(" << pid[i] << ")" << std::endl;
-		if (waitpid(pid[i], &wstatus, 0) != pid[i])
-			perror("ERROR: dkg-timestamp (waitpid)");
-		if (!WIFEXITED(wstatus))
-		{
-			std::cerr << "ERROR: protocol instance ";
-			if (WIFSIGNALED(wstatus))
-				std::cerr << pid[i] << " terminated by signal " <<
-					WTERMSIG(wstatus) << std::endl;
-			if (WCOREDUMP(wstatus))
-				std::cerr << pid[i] << " dumped core" << std::endl;
-			ret = -1; // fatal error
-		}
-		else if (WIFEXITED(wstatus))
-		{
-			if (opt_verbose)
-				std::cerr << "INFO: protocol instance " << pid[i] <<
-					" terminated with exit status " << WEXITSTATUS(wstatus) <<
-					std::endl;
-			if (WEXITSTATUS(wstatus))
-				ret = -2; // error
-		}
-		for (size_t j = 0; j < peers.size(); j++)
-		{
-			if ((close(pipefd[i][j][0]) < 0) || (close(pipefd[i][j][1]) < 0))
-				perror("ERROR: dkg-timestamp (close)");
-			if ((close(broadcast_pipefd[i][j][0]) < 0) ||
-				(close(broadcast_pipefd[i][j][1]) < 0))
-			{
-				perror("ERROR: dkg-timestamp (close)");
-			}
-		}
-	}
-	
+	if (should_unlock)
+		unlock_memory();
 	return ret;
 }
 
