@@ -1,7 +1,7 @@
 /*******************************************************************************
    This file is part of Distributed Privacy Guard (DKGPG).
 
- Copyright (C) 2017, 2018  Heiko Stamer <HeikoStamer@gmx.net>
+ Copyright (C) 2017, 2018, 2019  Heiko Stamer <HeikoStamer@gmx.net>
 
    DKGPG is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -326,22 +326,19 @@ void xtest
 }
 
 time_t agree_time
-	(const time_t sigtime,
+	(const time_t mytime,
 	 const size_t whoami,
 	 const size_t peers,
 	 const int opt_verbose,
 	 CachinKursawePetzoldShoupRBC *rbc)
 {
 	if (opt_verbose)
-	{
-		std::cerr << "INFO: agree on a signature creation time for OpenPGP" <<
-			std::endl;
-	}
+		std::cerr << "INFO: agree on a creation time for OpenPGP" << std::endl;
 	std::vector<time_t> tvs;
 	mpz_t mtv;
-	mpz_init_set_ui(mtv, sigtime);
+	mpz_init_set_ui(mtv, mytime);
 	rbc->Broadcast(mtv);
-	tvs.push_back(sigtime);
+	tvs.push_back(mytime);
 	for (size_t i = 0; i < peers; i++)
 	{
 		if (i != whoami)
@@ -354,8 +351,8 @@ time_t agree_time
 			}
 			else
 			{
-				std::cerr << "WARNING: p_" << whoami << ": no signature" <<
-					" creation time stamp received from p_" << i << std::endl;
+				std::cerr << "WARNING: p_" << whoami << ": no creation" <<
+					" timestamp received from p_" << i << std::endl;
 			}
 		}
 	}
@@ -370,16 +367,34 @@ time_t agree_time
 	{
 		std::cerr << "ERROR: p_" << whoami << ": no timestamps received" <<
 			std::endl;
-		tvs.push_back(0); // add a dummy return value
+		tvs.push_back(0); // add at least one dummy return value
 	}
 	// use a median value as some kind of gentle agreement
-	time_t csigtime = tvs[tvs.size()/2];
+	time_t coctime = tvs[tvs.size()/2];
 	if (opt_verbose)
 	{
-		std::cerr << "INFO: p_" << whoami << ": canonicalized signature" <<
-			" creation time = " << csigtime << std::endl;
+		std::cerr << "INFO: p_" << whoami << ": canonicalized OpenPGP" <<
+			" creation time = " << coctime << std::endl;
 	}
-	return csigtime;
+	return coctime;
+}
+
+bool select_hashalgo
+	(CanettiGennaroJareckiKrawczykRabinDSS *dss,
+	 tmcg_openpgp_hashalgo_t &hashalgo)
+{
+	if (dss == NULL)
+		return false;
+	// select hash algorithm for OpenPGP based on |q| (size in bit)
+	if (mpz_sizeinbase(dss->q, 2L) == 256)
+		hashalgo = TMCG_OPENPGP_HASHALGO_SHA256; // SHA256 (alg 8)
+	else if (mpz_sizeinbase(dss->q, 2L) == 384)
+		hashalgo = TMCG_OPENPGP_HASHALGO_SHA384; // SHA384 (alg 9)
+	else if (mpz_sizeinbase(dss->q, 2L) == 512)
+		hashalgo = TMCG_OPENPGP_HASHALGO_SHA512; // SHA512 (alg 10)
+	else
+		return false;
+	return true;
 }
 
 bool sign_hash
@@ -538,3 +553,99 @@ bool sign_hash
 	return true;
 }
 
+void canonicalize
+	(std::vector<std::string> &p)
+{
+	std::sort(p.begin(), p.end());
+	std::vector<std::string>::iterator it = std::unique(p.begin(), p.end());
+	p.resize(std::distance(p.begin(), it));
+}
+
+int wait_instance
+	(const size_t whoami,
+	 const int opt_verbose,
+	 pid_t pid[DKGPG_MAX_N])
+{
+	int wstatus = 0;
+	if (opt_verbose)
+		std::cerr << "INFO: waitpid(" << pid[whoami] << ")" << std::endl;
+	if (waitpid(pid[whoami], &wstatus, 0) != pid[whoami])
+	{
+		perror("ERROR: dkg-common:wait_instance (waitpid)");
+		return -1; // waitpid failed
+	}
+	if (!WIFEXITED(wstatus))
+	{
+		std::cerr << "ERROR: protocol instance ";
+		if (WIFSIGNALED(wstatus))
+		{
+			std::cerr << pid[whoami] << " terminated by signal " <<
+				WTERMSIG(wstatus) << std::endl;
+		}
+		if (WCOREDUMP(wstatus))
+			std::cerr << pid[whoami] << " dumped core" << std::endl;
+		return -1; // fatal error
+	}
+	else if (WIFEXITED(wstatus))
+	{
+		if (opt_verbose)
+		{
+			std::cerr << "INFO: protocol instance " << pid[whoami] <<
+				" terminated with exit status " << WEXITSTATUS(wstatus) <<
+				std::endl;
+		}
+		if (WEXITSTATUS(wstatus))
+			return -2; // error
+	}
+	return 0;
+}
+
+int run_localtest
+	(const size_t peers,
+	 const int opt_verbose,
+	 pid_t pid[DKGPG_MAX_N],
+	 int pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2],
+	 int bpipefd[DKGPG_MAX_N][DKGPG_MAX_N][2],
+	 void (*fork_instance)(const size_t))
+{
+	assert(peers <= DKGPG_MAX_N);
+	int ret = 0;
+	std::cerr << "WARNING: running only a local test with " << peers <<
+		" participants" << std::endl;
+	// open pipes
+	for (size_t i = 0; i < peers; i++)
+	{
+		for (size_t j = 0; j < peers; j++)
+		{
+			if (pipe(pipefd[i][j]) < 0)
+				perror("ERROR: dkg-common:run_localtest (pipe)");
+			if (pipe(bpipefd[i][j]) < 0)
+				perror("ERROR: dkg-common:run_localtest (pipe)");
+		}
+	}
+	
+	// start childs
+	for (size_t i = 0; i < peers; i++)
+		fork_instance(i);
+
+	// sleep for five seconds
+	sleep(5);
+	
+	// wait for childs and close pipes
+	for (size_t i = 0; i < peers; i++)
+	{
+		int iret = wait_instance(i, opt_verbose, pid);
+		if (iret != 0)
+			ret = iret; // return error, if any instance failed
+		for (size_t j = 0; j < peers; j++)
+		{
+			if ((close(pipefd[i][j][0]) < 0) || (close(pipefd[i][j][1]) < 0))
+				perror("ERROR: dkg-common:run_localtest (close)");
+			if ((close(bpipefd[i][j][0]) < 0) || (close(bpipefd[i][j][1]) < 0))
+				perror("ERROR: dkg-common:run_localtest (close)");
+		}
+	}
+	return ret;
+}
+
+
