@@ -1,7 +1,7 @@
 /*******************************************************************************
    This file is part of Distributed Privacy Guard (DKGPG).
 
- Copyright (C) 2017, 2018  Heiko Stamer <HeikoStamer@gmx.net>
+ Copyright (C) 2017, 2018, 2019  Heiko Stamer <HeikoStamer@gmx.net>
 
    DKGPG is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -28,6 +28,9 @@
 #ifdef DKGPG_TESTSUITE_Y
 	#undef GNUNET
 #endif
+#ifdef DKGPG_TESTSUITE_TS
+	#undef GNUNET
+#endif
 
 // copy infos from DKGPG package before overwritten by GNUnet headers
 static const char *version = PACKAGE_VERSION " (" PACKAGE_NAME ")";
@@ -51,6 +54,7 @@ static const char *protocol = "DKGPG-generate-1.0";
 #include "dkg-tcpip-common.hh"
 #include "dkg-gnunet-common.hh"
 #include "dkg-io.hh"
+#include "dkg-common.hh"
 
 int							pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2];
 int							broadcast_pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2];
@@ -62,9 +66,11 @@ tmcg_openpgp_secure_string_t	passphrase;
 std::string						userid, passwords, hostname, port;
 int 							opt_verbose = 0;
 bool							opt_y = false, opt_timestamping = false;
+bool							opt_nopassphrase = false;
 char							*opt_crs = NULL;
 char							*opt_passwords = NULL;
 char							*opt_hostname = NULL;
+char							*opt_u = NULL;
 unsigned long int				opt_t = DKGPG_MAX_N, opt_s = DKGPG_MAX_N;
 unsigned long int				opt_e = 0, opt_p = 55000, opt_W = 5;
 
@@ -113,8 +119,6 @@ void run_instance
 			hashalgo = TMCG_OPENPGP_HASHALGO_SHA384; // SHA384 (alg 9)
 		else if (mpz_sizeinbase(vtmf->q, 2L) == 512)
 			hashalgo = TMCG_OPENPGP_HASHALGO_SHA512; // SHA512 (alg 10)
-//		else if (mpz_sizeinbase(vtmf->q, 2L) > 1024)
-//			hashalgo = TMCG_OPENPGP_HASHALGO_SHA512; // SHA512 (alg 10) FIXME: we need dynamic hashsize here! SHAKE
 		else
 		{
 			std::cerr << "ERROR: selecting hash algorithm failed for |q| = " <<
@@ -223,7 +227,10 @@ void run_instance
 		CallasDonnerhackeFinneyShawThayerRFC4880::PacketUidEncode(userid, uid);
 		dsaflags.push_back(0x01 | 0x02);
 		if (opt_timestamping)
+		{
+			// FIXME: openpgp-wg: the first octet of usage flags should be zero?
 			dsaflags.push_back(0x08); // "This key may be used for timestamping"
+		}
 		sigtime = time(NULL); // current time
 		CallasDonnerhackeFinneyShawThayerRFC4880::
 			PacketSigPrepareSelfSignature(TMCG_OPENPGP_SIGNATURE_POSITIVE_CERTIFICATION,
@@ -505,24 +512,8 @@ void run_instance
 		new CachinKursawePetzoldShoupRBC(peers.size(), T_RBC, whoami, aiou2,
 		aiounicast::aio_scheduler_roundrobin, (opt_W * 60));
 	rbc->setID(myID);
-
 	// perform a simple exchange test with debug output
-	for (size_t i = 0; i < num_xtests; i++)
-	{
-		mpz_t xtest;
-		mpz_init_set_ui(xtest, i);
-		std::cerr << "INFO: P_" << whoami << ": xtest = " << xtest << " <-> ";
-		rbc->Broadcast(xtest);
-		for (size_t ii = 0; ii < peers.size(); ii++)
-		{
-			if (!rbc->DeliverFrom(xtest, ii))
-				std::cerr << "<X> ";
-			else
-				std::cerr << xtest << " ";
-		}
-		std::cerr << std::endl;
-		mpz_clear(xtest);
-	}
+	xtest(num_xtests, whoami, peers.size(), rbc);
 			
 	// create and exchange temporary VTMF keys in order to bootstrap the
 	// $h$-generation for tElG and tDSS/DSA protocols [JL00]
@@ -670,60 +661,14 @@ void run_instance
 
 	// all participants must agree on a common key creation time (OpenPGP),
 	// because otherwise key ID and subkey ID does not match
-	if (opt_verbose)
-		std::cerr << "INFO: agree on a key creation time for OpenPGP" <<
-			std::endl;
-	time_t ckeytime = 0;
-	std::vector<time_t> tvs;
-	mpz_t mtv;
-	mpz_init_set_ui(mtv, keytime);
-	rbc->Broadcast(mtv);
-	tvs.push_back(keytime);
-	for (size_t i = 0; i < peers.size(); i++)
-	{
-		if (i != whoami)
-		{
-			if (rbc->DeliverFrom(mtv, i))
-			{
-				time_t utv;
-				utv = (time_t)mpz_get_ui(mtv);
-				tvs.push_back(utv);
-			}
-			else
-			{
-				std::cerr << "WARNING: P_" << whoami << ": no key creation" <<
-					" time received from " << i << std::endl;
-			}
-		}
-	}
-	mpz_clear(mtv);
-	std::sort(tvs.begin(), tvs.end());
-	if (tvs.size() < (peers.size() - T_RBC))
-	{
-		std::cerr << "ERROR: P_" << whoami << ": not enough timestamps" <<
-			" received" << std::endl;
-		delete dkg, delete dss;
-		delete rbc, delete vtmf, delete aiou, delete aiou2;
-		exit(-1);
-	}
-	// use a median value as some kind of gentle agreement
-	ckeytime = tvs[tvs.size()/2];
-	if (opt_verbose)
-		std::cerr << "INFO: P_" << whoami << ": canonicalized key creation" <<
-			" time = " << ckeytime << std::endl;
+	time_t ckeytime = agree_time(keytime, whoami, peers.size(), opt_verbose, rbc);
 
 	// select hash algorithm for OpenPGP based on |q| (size in bit)
 	tmcg_openpgp_hashalgo_t hashalgo = TMCG_OPENPGP_HASHALGO_UNKNOWN;
-	if (mpz_sizeinbase(vtmf->q, 2L) == 256)
-		hashalgo = TMCG_OPENPGP_HASHALGO_SHA256; // SHA256 (alg 8)
-	else if (mpz_sizeinbase(vtmf->q, 2L) == 384)
-		hashalgo = TMCG_OPENPGP_HASHALGO_SHA384; // SHA384 (alg 9)
-	else if (mpz_sizeinbase(vtmf->q, 2L) == 512)
-		hashalgo = TMCG_OPENPGP_HASHALGO_SHA512; // SHA512 (alg 10)
-	else
+	if (!select_hashalgo(dss, hashalgo))
 	{
 		std::cerr << "ERROR: P_" << whoami << ": selecting hash algorithm" <<
-			" failed for |q| = " << mpz_sizeinbase(vtmf->q, 2L) << std::endl;
+			" failed for |q| = " << mpz_sizeinbase(dss->q, 2L) << std::endl;
 		delete dkg, delete dss;
 		delete rbc, delete vtmf, delete aiou, delete aiou2;
 		exit(-1);
@@ -1030,7 +975,10 @@ void run_instance
 		// split by a secret-sharing mechanism
 		dsaflags.push_back(0x01 | 0x02 | 0x10);
 		if (opt_timestamping)
+		{
+			// FIXME: openpgp-wg: the first octet of usage flags should be zero?
 			dsaflags.push_back(0x08); // "This key may be used for timestamping"
+		}
 		// reuse key creation time as signature creation time
 		sigtime = ckeytime;
 	}
@@ -1039,7 +987,10 @@ void run_instance
 		// key may be used to certify other keys and to sign data
 		dsaflags.push_back(0x01 | 0x02);
 		if (opt_timestamping)
+		{
+			// FIXME: openpgp-wg: the first octet of usage flags should be zero?
 			dsaflags.push_back(0x08); // "This key may be used for timestamping"
+		}
 		// for a non-shared DSA primary key no common timestamp required 
 		sigtime = time(NULL); // current time
 	}
@@ -1537,7 +1488,7 @@ void run_instance
 	gcry_sexp_release(key);
 	
 	// at the end: deliver some more rounds for still waiting parties
-	time_t synctime = aiounicast::aio_timeout_long;
+	time_t synctime = (opt_W * 6);
 	if (opt_verbose)
 		std::cerr << "INFO: P_" << whoami << ": waiting approximately " <<
 			(synctime * (T_RBC + 1)) << " seconds for stalled parties" <<
@@ -1881,11 +1832,159 @@ bool fips_verify
 		return true;
 }
 
+bool pqg_extract
+	(std::string crs, const bool fips,
+	 mpz_ptr fips_p, mpz_ptr fips_q, mpz_ptr fips_g)
+{
+	std::string mpz_str;
+	mpz_t crsmpz;
+	mpz_init(crsmpz);
+	for (size_t i = 0; i < 4; i++)
+	{
+		if (!TMCG_ParseHelper::gs(crs, '|', mpz_str))
+		{
+			std::cerr << "ERROR: common reference string (CRS) is corrupted" <<
+				std::endl;
+			mpz_clear(crsmpz);
+			return false;
+		}
+		else if ((mpz_set_str(crsmpz, mpz_str.c_str(), TMCG_MPZ_IO_BASE) < 0) ||
+			!TMCG_ParseHelper::nx(crs, '|'))
+		{
+			std::cerr << "ERROR: common reference string (CRS) is corrupted" <<
+				std::endl;
+			mpz_clear(crsmpz);
+			return false;
+		}
+		crss << crsmpz << std::endl;
+		if (i == 0)
+		{
+			mpz_set(fips_p, crsmpz);
+			if (opt_verbose > 1)
+				std::cerr << "INFO: p";
+		}
+		else if (i == 1)
+		{
+			mpz_set(fips_q, crsmpz);
+			if (opt_verbose > 1)
+				std::cerr << "INFO: q";
+		}
+		else if (i == 2)
+		{
+			mpz_set(fips_g, crsmpz);
+			if (opt_verbose > 1)
+				std::cerr << "INFO: g";
+		}
+		if ((opt_verbose > 1) && (i < 3))
+			std::cerr << " (" << mpz_sizeinbase(crsmpz, 2L) << " bits) = " <<
+				crsmpz << std::endl;
+	}
+	mpz_clear(crsmpz);
+	if (fips)
+	{
+		mpz_t fips_hashalgo, fips_dps, fips_counter, fips_index;
+		mpz_init_set_ui(fips_hashalgo, 0L), mpz_init_set_ui(fips_dps, 0L);
+		mpz_init_set_ui(fips_counter, 0L), mpz_init_set_ui(fips_index, 0L);
+		if (!TMCG_ParseHelper::gs(crs, '|', mpz_str))
+		{
+			std::cerr << "ERROR: common reference string (CRS) is corrupted" <<
+				std::endl;
+			mpz_clear(fips_hashalgo), mpz_clear(fips_dps);
+			mpz_clear(fips_counter), mpz_clear(fips_index);
+			return false;
+		}
+		if ((mpz_set_str(fips_hashalgo, mpz_str.c_str(),
+			TMCG_MPZ_IO_BASE) < 0) || !TMCG_ParseHelper::nx(crs, '|'))
+		{
+			std::cerr << "ERROR: common reference string (CRS) is corrupted" <<
+				std::endl;
+			mpz_clear(fips_hashalgo), mpz_clear(fips_dps);
+			mpz_clear(fips_counter), mpz_clear(fips_index);
+			return false;
+		}
+		if (!TMCG_ParseHelper::gs(crs, '|', mpz_str))
+		{
+			std::cerr << "ERROR: common reference string (CRS) is corrupted" <<
+				std::endl;
+			mpz_clear(fips_hashalgo), mpz_clear(fips_dps);
+			mpz_clear(fips_counter), mpz_clear(fips_index);
+			return false;
+		}
+		if ((mpz_set_str(fips_dps, mpz_str.c_str(), TMCG_MPZ_IO_BASE) < 0) ||
+			(!TMCG_ParseHelper::nx(crs, '|')))
+		{
+			std::cerr << "ERROR: common reference string (CRS) is corrupted" <<
+				std::endl;
+			mpz_clear(fips_hashalgo), mpz_clear(fips_dps);
+			mpz_clear(fips_counter), mpz_clear(fips_index);
+			return false;
+		}
+		if (!TMCG_ParseHelper::gs(crs, '|', mpz_str))
+		{
+			std::cerr << "ERROR: common reference string (CRS) is corrupted" <<
+				std::endl;
+			mpz_clear(fips_hashalgo), mpz_clear(fips_dps);
+			mpz_clear(fips_counter), mpz_clear(fips_index);
+			return false;
+		}
+		if ((mpz_set_str(fips_counter, mpz_str.c_str(),
+			TMCG_MPZ_IO_BASE) < 0) || (!TMCG_ParseHelper::nx(crs, '|')))
+		{
+			std::cerr << "ERROR: common reference string (CRS) is corrupted" <<
+				std::endl;
+			mpz_clear(fips_hashalgo), mpz_clear(fips_dps);
+			mpz_clear(fips_counter), mpz_clear(fips_index);
+			return false;
+		}
+		if (!TMCG_ParseHelper::gs(crs, '|', mpz_str))
+		{
+			std::cerr << "ERROR: common reference string (CRS) is corrupted" <<
+				std::endl;
+			mpz_clear(fips_hashalgo), mpz_clear(fips_dps);
+			mpz_clear(fips_counter), mpz_clear(fips_index);
+			return false;
+		}
+		if ((mpz_set_str(fips_index, mpz_str.c_str(), TMCG_MPZ_IO_BASE) < 0) ||
+			(!TMCG_ParseHelper::nx(crs, '|')))
+		{
+			std::cerr << "ERROR: common reference string (CRS) is corrupted" <<
+				std::endl;
+			mpz_clear(fips_hashalgo), mpz_clear(fips_dps);
+			mpz_clear(fips_counter), mpz_clear(fips_index);
+			return false;
+		}
+		if (mpz_get_ui(fips_hashalgo) != GCRY_MD_SHA256) 
+		{
+			std::cerr << "ERROR: hash function is not approved according to" <<
+				" FIPS 186-4" << std::endl;
+			mpz_clear(fips_hashalgo), mpz_clear(fips_dps);
+			mpz_clear(fips_counter), mpz_clear(fips_index);
+			return false;
+		}
+		// check the domain parameters according to FIPS 186-4 sections
+		// A.1.1.3 and A.2.4
+		if (!fips_verify(fips_p, fips_q, fips_g, fips_hashalgo, fips_dps,
+			fips_counter, fips_index))
+		{
+			std::cerr << "ERROR: domain parameters are not set according to" <<
+				" FIPS 186-4" << std::endl;
+			mpz_clear(fips_hashalgo), mpz_clear(fips_dps);
+			mpz_clear(fips_counter), mpz_clear(fips_index);
+			return false;
+		}
+		// release
+		mpz_clear(fips_hashalgo), mpz_clear(fips_dps);
+		mpz_clear(fips_counter), mpz_clear(fips_index);
+	}
+	return true;
+}
+
 #ifdef GNUNET
 char *gnunet_opt_crs = NULL;
 char *gnunet_opt_hostname = NULL;
 char *gnunet_opt_passwords = NULL;
 char *gnunet_opt_port = NULL;
+char *gnunet_opt_u = NULL;
 unsigned int gnunet_opt_t_resilience = DKGPG_MAX_N;
 unsigned int gnunet_opt_s_resilience = DKGPG_MAX_N;
 unsigned int gnunet_opt_keyexptime = 0;
@@ -1895,6 +1994,7 @@ unsigned int gnunet_opt_W = opt_W;
 int gnunet_opt_verbose = 0;
 int gnunet_opt_y = 0;
 int gnunet_opt_timestamping = 0;
+int gnunet_opt_nopassphrase = 0;
 #endif
 
 void fork_instance
@@ -1934,7 +2034,7 @@ int main
 	// setup CRS (common reference string) |p| = 3072 bit, |q| = 256 bit
 	// The domain parameters of the underlying fixed group have been
 	// generated by the author. You can run dkg-gencrs and then use the
-	// option "-g" to employ your own domain parameter set.
+	// option "-g" to employ your own individual domain parameter set.
 	static std::string crs = "fips-crs|etc0k13hu6mDWyye0MUepXvXJf1M6Uqt13"
 		"mAyAEhJ292TMOxG5HilFlFaG2YUIXsPISxWKjZqBY0VO6YjF1DIGsIzQ6GK9"
 		"Myk9MGSVxP44lfjVIodRRFk2anRLktvO2Kcq0Z6hP3yvQNmM0sl5JTvzrJLt"
@@ -1991,6 +2091,11 @@ int main
 		),
 		GNUNET_GETOPT_option_logfile(&logfile),
 		GNUNET_GETOPT_option_loglevel(&loglev),
+		GNUNET_GETOPT_option_flag('N',
+			"no-passphrase",
+			"disable private key protection",
+			&gnunet_opt_nopassphrase
+		),
 		GNUNET_GETOPT_option_string('p',
 			"port",
 			"STRING",
@@ -2020,6 +2125,12 @@ int main
 			"state that the generated key is used for timestamping",
 			&gnunet_opt_timestamping
 		),
+		GNUNET_GETOPT_option_string('u',
+			"uid",
+			"STRING",
+			"user ID of the generated key",
+			&gnunet_opt_u
+		),
 		GNUNET_GETOPT_option_version(version),
 		GNUNET_GETOPT_option_flag('V',
 			"verbose",
@@ -2076,6 +2187,11 @@ int main
 		hostname = gnunet_opt_hostname; // get hostname from GNUnet options
 	if (gnunet_opt_W != opt_W)
 		opt_W = gnunet_opt_W; // get aiou message timeout from GNUnet options
+	if (gnunet_opt_u != NULL)
+	{
+		userid = gnunet_opt_u; // get userid from GNUnet options
+		opt_u = gnunet_opt_u;
+	}
 #endif
 
 	// create peer list from remaining arguments
@@ -2089,7 +2205,7 @@ int main
 			(arg.find("-l") == 0) || (arg.find("-g") == 0) ||
 			(arg.find("-x") == 0) || (arg.find("-s") == 0) ||
 			(arg.find("-e") == 0) || (arg.find("-P") == 0) ||
-			(arg.find("-H") == 0))
+			(arg.find("-H") == 0) || (arg.find("-u") == 0))
 		{
 			size_t idx = ++i;
 			if ((arg.find("-g") == 0) && (idx < (size_t)(argc - 1)) &&
@@ -2130,6 +2246,12 @@ int main
 			{
 				port = argv[i+1];
 			}
+			if ((arg.find("-u") == 0) && (idx < (size_t)(argc - 1)) &&
+				(opt_u == NULL))
+			{
+				userid = argv[i+1];
+				opt_u = (char*)userid.c_str();
+			}
 			if ((arg.find("-W") == 0) && (idx < (size_t)(argc - 1)) &&
 				(opt_W == 5))
 			{
@@ -2139,7 +2261,8 @@ int main
 		}
 		else if ((arg.find("--") == 0) || (arg.find("-v") == 0) ||
 			(arg.find("-h") == 0) || (arg.find("-V") == 0) ||
-			(arg.find("-y") == 0) || (arg.find("-T") == 0))
+			(arg.find("-y") == 0) || (arg.find("-T") == 0) ||
+			(arg.find("-N") == 0))
 		{
 			if ((arg.find("-h") == 0) || (arg.find("--help") == 0))
 			{
@@ -2155,6 +2278,8 @@ int main
 					" defines underlying DDH-hard group" << std::endl;
 				std::cout << "  -H STRING      hostname (e.g. onion address)" <<
 					" of this peer within PEERS" << std::endl;
+				std::cout << "  -N, --no-passphrase  disable private key" <<
+					" protection" << std::endl;
 				std::cout << "  -p INTEGER     start port for built-in" <<
 					" TCP/IP message exchange service" << std::endl; 
 				std::cout << "  -P STRING      exchanged passwords to" <<
@@ -2163,8 +2288,10 @@ int main
 					" protocol (signature scheme)" << std::endl;
 				std::cout << "  -t INTEGER     resilience of tElG protocol" <<
 					" (threshold decryption)" << std::endl;
-				std::cout << "  -T, --timestamping state that the generated" <<
+				std::cout << "  -T, --timestamping  state that the generated" <<
 					" key is used for timestamping" << std::endl;
+				std::cout << "  -u STRING      user ID of the generated key" <<
+					std::endl;
 				std::cout << "  -v, --version  print the version number" <<
 					std::endl;
 				std::cout << "  -V, --verbose  turn on verbose output" <<
@@ -2190,6 +2317,8 @@ int main
 				opt_y = true;
 			if ((arg.find("-T") == 0) || (arg.find("--timestamping") == 0))
 				opt_timestamping = true;
+			if ((arg.find("-N") == 0) || (arg.find("--no-passphrase") == 0))
+				opt_nopassphrase = true;
 			continue;
 		}
 		else if (arg.find("-") == 0)
@@ -2209,6 +2338,7 @@ int main
 			return -1;
 		}
 	}
+
 #ifdef DKGPG_TESTSUITE
 	peers.push_back("Test1");
 	peers.push_back("Test2");
@@ -2216,8 +2346,6 @@ int main
 	peers.push_back("Test4");
 	opt_verbose = 2;
 	opt_e = 10800;
-	if (tmcg_mpz_wrandom_ui() % 2)
-		opt_timestamping = true;
 	if (tmcg_mpz_wrandom_ui() % 2)
 	{
 		// sometimes test a non-FIPS CRS
@@ -2242,38 +2370,22 @@ int main
 			"0000000000000000000000000000000000000000000000000000"
 			"0000000000000000000000000000000000000000000000000|";
 	}
-//	else if (tmcg_mpz_wrandom_ui() % 2)
-//	{
-//		// sometimes test a CRS with fixed parameters from RFC 7919
-//		crs = "rfc-crs|m88HxBJpj9yKzDAlXe8tQ0jP0zSNm7oE71YSFt2ZQHiLVO"
-//			"kqnFJOK5EcaSLRHGmrtAhVybt7MarfGJf9DP5NTig8GXokeeHMiR"
-//			"00pxXnWmMmNd4C4JbsbmSKfOyPai66hMPXDybZprvk6E9Towmh8s"
-//			"rAAlz36OkTVJrZ5cKSEnpnDdd0BBwzvGDlstU5ZcuVwdVB7us5tX"
-//			"zyucSJBYoCEWf7VvqbJVYhe4GdvCczMCHGUsEKMsXsg5oBRHyQ6K"
-//			"TXX6PCJp9eo2kB4OvHgI5sZS0FMtzpuDlwcqYdLowuRIucwCFp9o"
-//			"PSBFSvO9TZRMcRXorL6mZQBglR9CcLSg7OSd2Vaa333uj1WH6HfB"
-//			"tkIaRzfwgn1P2XUGQZotOci8gdrvijNFyIdJE6k6qoVMog4SoNHo"
-//			"HZrq1CHx1QqfM2E0bXDAQNCkgzQO7zvcteMCn7xHBwNzQBg0AabM"
-//			"34izE4VrBwpLK3dNTos1CDmDCtTOQ7S0DHYOmSkkKYAmSWMf2bot"
-//			"Yd|O448yaeurZzAUbaNlp4Ri0MhVUjBt3u73VmE7wWHi8rAkhNQO"
-//			"cehA2cJIEAiddOQwaLkzIwYgIQpd9pZbhXgjrL48GuNKK8gMDV0P"
-//			"yltlOBOBoX629nwItEAKhUCnM33LgClbzImuvxs374juTOLZRQa5"
-//			"NzWYCNEkevmXoAE7OutbooV5ayUxd6swRk2moSFyJkaYxR2wlzzS"
-//			"JE9amP67GKYkxvIekmLp28JxbJUg68dFR7ABRGwL2u5idzD3AEll"
-//			"YCb9uZpP1N5XCSdq92wHj07gRzux6syJQHJfuTSDeSJT67uZuCj5"
-//			"cjSh4jmigJDluQfYOHi5qNiZbJAjL3hEJWFnI1WWxMVl8Y8paws9"
-//			"IDzpyLOVhWGk8DHuRhJM4LJvxrMgcz9Jec3N3QPFgPL2EPBdu8mv"
-//			"v0b8yViQKg170IlbaDBbNLUiC3zxoRpB6OYydayBzi5q05IIg1XM"
-//			"Uc2FvayPff1ogjuR0b6t6bRjhD3j06dmCOENNAH5OEGBKWIuRmJ|2|1|";
-//	}
+#else
+#ifdef DKGPG_TESTSUITE_TS
+	peers.push_back("TestTS1");
+	peers.push_back("TestTS2");
+	peers.push_back("TestTS3");
+	peers.push_back("TestTS4");
+	opt_verbose = 2;
+	opt_e = 10800;
+	opt_timestamping = true;
 #else
 #ifdef DKGPG_TESTSUITE_Y
 	peers.push_back("TestY");
 	opt_y = true;
 	opt_verbose = 2;
 	opt_e = 10800;
-	if (tmcg_mpz_wrandom_ui() % 2)
-		opt_timestamping = true;
+	opt_timestamping = true;
 	if (tmcg_mpz_wrandom_ui() % 2)
 	{
 		// sometimes test a non-FIPS CRS
@@ -2298,30 +2410,7 @@ int main
 			"0000000000000000000000000000000000000000000000000000"
 			"0000000000000000000000000000000000000000000000000|";
 	}
-//	else if (tmcg_mpz_wrandom_ui() % 2)
-//	{
-//		// sometimes test a CRS with fixed parameters from RFC 7919
-//		crs = "rfc-crs|m88HxBJpj9yKzDAlXe8tQ0jP0zSNm7oE71YSFt2ZQHiLVO"
-//			"kqnFJOK5EcaSLRHGmrtAhVybt7MarfGJf9DP5NTig8GXokeeHMiR"
-//			"00pxXnWmMmNd4C4JbsbmSKfOyPai66hMPXDybZprvk6E9Towmh8s"
-//			"rAAlz36OkTVJrZ5cKSEnpnDdd0BBwzvGDlstU5ZcuVwdVB7us5tX"
-//			"zyucSJBYoCEWf7VvqbJVYhe4GdvCczMCHGUsEKMsXsg5oBRHyQ6K"
-//			"TXX6PCJp9eo2kB4OvHgI5sZS0FMtzpuDlwcqYdLowuRIucwCFp9o"
-//			"PSBFSvO9TZRMcRXorL6mZQBglR9CcLSg7OSd2Vaa333uj1WH6HfB"
-//			"tkIaRzfwgn1P2XUGQZotOci8gdrvijNFyIdJE6k6qoVMog4SoNHo"
-//			"HZrq1CHx1QqfM2E0bXDAQNCkgzQO7zvcteMCn7xHBwNzQBg0AabM"
-//			"34izE4VrBwpLK3dNTos1CDmDCtTOQ7S0DHYOmSkkKYAmSWMf2bot"
-//			"Yd|O448yaeurZzAUbaNlp4Ri0MhVUjBt3u73VmE7wWHi8rAkhNQO"
-//			"cehA2cJIEAiddOQwaLkzIwYgIQpd9pZbhXgjrL48GuNKK8gMDV0P"
-//			"yltlOBOBoX629nwItEAKhUCnM33LgClbzImuvxs374juTOLZRQa5"
-//			"NzWYCNEkevmXoAE7OutbooV5ayUxd6swRk2moSFyJkaYxR2wlzzS"
-//			"JE9amP67GKYkxvIekmLp28JxbJUg68dFR7ABRGwL2u5idzD3AEll"
-//			"YCb9uZpP1N5XCSdq92wHj07gRzux6syJQHJfuTSDeSJT67uZuCj5"
-//			"cjSh4jmigJDluQfYOHi5qNiZbJAjL3hEJWFnI1WWxMVl8Y8paws9"
-//			"IDzpyLOVhWGk8DHuRhJM4LJvxrMgcz9Jec3N3QPFgPL2EPBdu8mv"
-//			"v0b8yViQKg170IlbaDBbNLUiC3zxoRpB6OYydayBzi5q05IIg1XM"
-//			"Uc2FvayPff1ogjuR0b6t6bRjhD3j06dmCOENNAH5OEGBKWIuRmJ|2|1|";
-//	}
+#endif
 #endif
 #endif
 
@@ -2340,10 +2429,7 @@ int main
 	}
 
 	// canonicalize peer list and setup threshold values for tDSS/DSA and tElG
-	std::sort(peers.begin(), peers.end());
-	std::vector<std::string>::iterator it =
-		std::unique(peers.begin(), peers.end());
-	peers.resize(std::distance(peers.begin(), it));
+	canonicalize(peers);
 	T = (peers.size() - 1) / 2; // default: maximum t-resilience for tElG
 	S = (peers.size() - 1) / 2; // default: maximum s-resilience for tDSS/DSA
 	if (((peers.size() < 3)  || (peers.size() > DKGPG_MAX_N)) && !opt_y)
@@ -2364,7 +2450,7 @@ int main
 	}
 
 	// lock memory
-	bool force_secmem = false;
+	bool force_secmem = false, should_unlock = false;
 	if (!lock_memory())
 	{
 		std::cerr << "WARNING: locking memory failed; CAP_IPC_LOCK required" <<
@@ -2372,11 +2458,15 @@ int main
 		// at least try to use libgcrypt's secure memory
 		force_secmem = true;
 	}
+	else
+		should_unlock = true;
 
 	// initialize LibTMCG
 	if (!init_libTMCG(force_secmem))
 	{
 		std::cerr << "ERROR: initialization of LibTMCG failed" << std::endl;
+		if (should_unlock)
+			unlock_memory();
 		return -1;
 	}
 	if (opt_verbose)
@@ -2388,31 +2478,55 @@ int main
 	userid = "TestGroup <testing@localhost>";
 	passphrase = "Test";
 #else
+#ifdef DKGPG_TESTSUITE_TS
+	userid = "TestGroupTS <testing@localhost>";
+	passphrase = "Test";
+#else
 #ifdef DKGPG_TESTSUITE_Y
 	userid = "TestGroupY <testing@localhost>";
 	passphrase = "TestY";
 #else
-	std::cerr << "1. Please enter an OpenPGP-style user ID (name <email>): ";
-	std::getline(std::cin, userid);
-	std::cin.clear();
-	tmcg_openpgp_secure_string_t passphrase_check;
-	std::string ps1 = "2. Passphrase to protect your part of the private key";
-	std::string ps2 = "Please repeat the given passphrase to continue";
-	do
+	if (userid.length() == 0)
 	{
-		passphrase = "", passphrase_check = "";
-		if (!get_passphrase(ps1, false, passphrase))
-			return -1;
-		if (!get_passphrase(ps2, false, passphrase_check))
-			return -1;
-		if (passphrase != passphrase_check)
-			std::cerr << "WARNING: passphrase does not match;" <<
-				" please try again" << std::endl;
-		else if (passphrase == "")
-			std::cerr << "WARNING: no key protection due to empty passphrase" <<
-				std::endl;
+		std::cerr << "Please enter an OpenPGP-style user ID (name <email>): ";
+		std::getline(std::cin, userid);
+		std::cin.clear();
 	}
-	while (passphrase != passphrase_check);
+	if (opt_nopassphrase)
+	{
+		std::cerr << "WARNING: private key protection disabled due to option" <<
+			" --no-passphrase" << std::endl;
+	}
+	else
+	{
+		tmcg_openpgp_secure_string_t passphrase_check;
+		std::string ps1 = "Passphrase to protect your part of the private key";
+		std::string ps2 = "Please repeat the given passphrase to continue";
+		do
+		{
+			passphrase = "", passphrase_check = "";
+			if (!get_passphrase(ps1, false, passphrase))
+			{
+				if (should_unlock)
+					unlock_memory();
+				return -1;
+			}
+			if (!get_passphrase(ps2, false, passphrase_check))
+			{
+				if (should_unlock)
+					unlock_memory();
+				return -1;
+			}
+			if (passphrase != passphrase_check)
+				std::cerr << "WARNING: passphrase does not match;" <<
+					" please try again" << std::endl;
+			else if (passphrase == "")
+				std::cerr << "WARNING: private key protection disabled due" <<
+					" to empty passphrase" << std::endl;
+		}
+		while (passphrase != passphrase_check);
+	}
+#endif
 #endif
 #endif
 
@@ -2431,7 +2545,7 @@ int main
 		T = (peers.size() - 1); // apply an upper limit on T
 	if (S > ((peers.size() - 1) / 2))
 		S = (peers.size() - 1) / 2; // apply an upper limit on S
-	// check magic of CRS (common reference string)
+	// check magic bytes of CRS (common reference string)
 	if (TMCG_ParseHelper::cm(crs, "crs", '|'))
 	{
 		if (opt_verbose)
@@ -2456,166 +2570,24 @@ int main
 	{
 		std::cerr << "ERROR: common reference string (CRS) is not valid" <<
 			std::endl;
+		if (should_unlock)
+			unlock_memory();
 		return -1;
 	}
-	// parse p, q, g, k from CRS
-	std::string mpz_str;
-	mpz_t crsmpz, fips_p, fips_q, fips_g;
-	mpz_init(crsmpz);
+	// extract p, q, g from CRS
+	mpz_t fips_p, fips_q, fips_g;
 	mpz_init(fips_p), mpz_init(fips_q), mpz_init(fips_g);
-	for (size_t i = 0; i < 4; i++)
+	if (!pqg_extract(crs, fips, fips_p, fips_q, fips_g))
 	{
-		if (!TMCG_ParseHelper::gs(crs, '|', mpz_str))
-		{
-			std::cerr << "ERROR: common reference string (CRS) is corrupted" <<
-				std::endl;
-			mpz_clear(crsmpz);
-			mpz_clear(fips_p), mpz_clear(fips_q), mpz_clear(fips_g);
-			return -1;
-		}
-		else if ((mpz_set_str(crsmpz, mpz_str.c_str(), TMCG_MPZ_IO_BASE) < 0) ||
-			!TMCG_ParseHelper::nx(crs, '|'))
-		{
-			std::cerr << "ERROR: common reference string (CRS) is corrupted" <<
-				std::endl;
-			mpz_clear(crsmpz);
-			mpz_clear(fips_p), mpz_clear(fips_q), mpz_clear(fips_g);
-			return -1;
-		}
-		crss << crsmpz << std::endl;
-		if (i == 0)
-		{
-			mpz_set(fips_p, crsmpz);
-			if (opt_verbose > 1)
-				std::cerr << "INFO: p";
-		}
-		else if (i == 1)
-		{
-			mpz_set(fips_q, crsmpz);
-			if (opt_verbose > 1)
-				std::cerr << "INFO: q";
-		}
-		else if (i == 2)
-		{
-			mpz_set(fips_g, crsmpz);
-			if (opt_verbose > 1)
-				std::cerr << "INFO: g";
-		}
-		if ((opt_verbose > 1) && (i < 3))
-			std::cerr << " (" << mpz_sizeinbase(crsmpz, 2L) << " bits) = " <<
-				crsmpz << std::endl;
-	}
-	mpz_clear(crsmpz);
-	if (fips)
-	{
-		mpz_t fips_hashalgo, fips_dps, fips_counter, fips_index;
-		mpz_init_set_ui(fips_hashalgo, 0L), mpz_init_set_ui(fips_dps, 0L);
-		mpz_init_set_ui(fips_counter, 0L), mpz_init_set_ui(fips_index, 0L);
-		if (!TMCG_ParseHelper::gs(crs, '|', mpz_str))
-		{
-			std::cerr << "ERROR: common reference string (CRS) is corrupted" <<
-				std::endl;
-			mpz_clear(fips_p), mpz_clear(fips_q), mpz_clear(fips_g);
-			mpz_clear(fips_hashalgo), mpz_clear(fips_dps);
-			mpz_clear(fips_counter), mpz_clear(fips_index);
-			return -1;
-		}
-		if ((mpz_set_str(fips_hashalgo, mpz_str.c_str(),
-			TMCG_MPZ_IO_BASE) < 0) || !TMCG_ParseHelper::nx(crs, '|'))
-		{
-			std::cerr << "ERROR: common reference string (CRS) is corrupted" <<
-				std::endl;
-			mpz_clear(fips_p), mpz_clear(fips_q), mpz_clear(fips_g);
-			mpz_clear(fips_hashalgo), mpz_clear(fips_dps);
-			mpz_clear(fips_counter), mpz_clear(fips_index);
-			return -1;
-		}
-		if (!TMCG_ParseHelper::gs(crs, '|', mpz_str))
-		{
-			std::cerr << "ERROR: common reference string (CRS) is corrupted" <<
-				std::endl;
-			mpz_clear(fips_p), mpz_clear(fips_q), mpz_clear(fips_g);
-			mpz_clear(fips_hashalgo), mpz_clear(fips_dps);
-			mpz_clear(fips_counter), mpz_clear(fips_index);
-			return -1;
-		}
-		if ((mpz_set_str(fips_dps, mpz_str.c_str(), TMCG_MPZ_IO_BASE) < 0) ||
-			(!TMCG_ParseHelper::nx(crs, '|')))
-		{
-			std::cerr << "ERROR: common reference string (CRS) is corrupted" <<
-				std::endl;
-			mpz_clear(fips_p), mpz_clear(fips_q), mpz_clear(fips_g);
-			mpz_clear(fips_hashalgo), mpz_clear(fips_dps);
-			mpz_clear(fips_counter), mpz_clear(fips_index);
-			return -1;
-		}
-		if (!TMCG_ParseHelper::gs(crs, '|', mpz_str))
-		{
-			std::cerr << "ERROR: common reference string (CRS) is corrupted" <<
-				std::endl;
-			mpz_clear(fips_p), mpz_clear(fips_q), mpz_clear(fips_g);
-			mpz_clear(fips_hashalgo), mpz_clear(fips_dps);
-			mpz_clear(fips_counter), mpz_clear(fips_index);
-			return -1;
-		}
-		if ((mpz_set_str(fips_counter, mpz_str.c_str(),
-			TMCG_MPZ_IO_BASE) < 0) || (!TMCG_ParseHelper::nx(crs, '|')))
-		{
-			std::cerr << "ERROR: common reference string (CRS) is corrupted" <<
-				std::endl;
-			mpz_clear(fips_p), mpz_clear(fips_q), mpz_clear(fips_g);
-			mpz_clear(fips_hashalgo), mpz_clear(fips_dps);
-			mpz_clear(fips_counter), mpz_clear(fips_index);
-			return -1;
-		}
-		if (!TMCG_ParseHelper::gs(crs, '|', mpz_str))
-		{
-			std::cerr << "ERROR: common reference string (CRS) is corrupted" <<
-				std::endl;
-			mpz_clear(fips_p), mpz_clear(fips_q), mpz_clear(fips_g);
-			mpz_clear(fips_hashalgo), mpz_clear(fips_dps);
-			mpz_clear(fips_counter), mpz_clear(fips_index);
-			return -1;
-		}
-		if ((mpz_set_str(fips_index, mpz_str.c_str(), TMCG_MPZ_IO_BASE) < 0) ||
-			(!TMCG_ParseHelper::nx(crs, '|')))
-		{
-			std::cerr << "ERROR: common reference string (CRS) is corrupted" <<
-				std::endl;
-			mpz_clear(fips_p), mpz_clear(fips_q), mpz_clear(fips_g);
-			mpz_clear(fips_hashalgo), mpz_clear(fips_dps);
-			mpz_clear(fips_counter), mpz_clear(fips_index);
-			return -1;
-		}
-		if (mpz_get_ui(fips_hashalgo) != GCRY_MD_SHA256) 
-		{
-			std::cerr << "ERROR: hash function is not approved according to" <<
-				" FIPS 186-4" << std::endl;
-			mpz_clear(fips_p), mpz_clear(fips_q), mpz_clear(fips_g);
-			mpz_clear(fips_hashalgo), mpz_clear(fips_dps);
-			mpz_clear(fips_counter), mpz_clear(fips_index);
-			return -1;
-		}
-		// check the domain parameters according to FIPS 186-4 sections
-		// A.1.1.3 and A.2.4
-		if (!fips_verify(fips_p, fips_q, fips_g, fips_hashalgo, fips_dps,
-			fips_counter, fips_index))
-		{
-			std::cerr << "ERROR: domain parameters are not set according to" <<
-				" FIPS 186-4" << std::endl;
-			mpz_clear(fips_p), mpz_clear(fips_q), mpz_clear(fips_g);
-			mpz_clear(fips_hashalgo), mpz_clear(fips_dps);
-			mpz_clear(fips_counter), mpz_clear(fips_index);
-			return -1;
-		}
-		// release
-		mpz_clear(fips_hashalgo), mpz_clear(fips_dps);
-		mpz_clear(fips_counter), mpz_clear(fips_index);
+		mpz_clear(fips_p), mpz_clear(fips_q), mpz_clear(fips_g);
+		if (should_unlock)
+			unlock_memory();
+		return -1;
 	}
 	// initialize cache
-	std::cerr << "3. We need a lot of entropy to cache very strong" <<
+	std::cerr << "We need a lot of entropy to cache very strong" <<
 		" randomness for key generation." << std::endl;
-	std::cerr << "   Please use other programs, move the mouse, and type on" <<
+	std::cerr << "Please use other programs, move the mouse, and type on" <<
 		" your keyboard: " << std::endl;
 	if (opt_y)
 		tmcg_mpz_ssrandomm_cache_init(cache, cache_mod, cache_avail, 2, fips_q);
@@ -2624,207 +2596,125 @@ int main
 			((2 * (S + 1)) + (2 * (T + 1))), fips_q);
 	std::cerr << "Thank you!" << std::endl;
 	mpz_clear(fips_p), mpz_clear(fips_q), mpz_clear(fips_g);
-	// initialize return code
+	// initialize return code and do the main work
 	int ret = 0;
-	// create underlying point-to-point channels, if built-in TCP/IP requested
 	if ((opt_hostname != NULL) && !opt_y)
 	{
-		if (port.length())
-			opt_p = strtoul(port.c_str(), NULL, 10); // get start port from "-p"
-		if ((opt_p < 1) || (opt_p > 65535))
-		{
-			std::cerr << "ERROR: no valid TCP start port given" << std::endl;
-			return -1;
-		}
-		tcpip_init(hostname);
-		tcpip_bindports((uint16_t)opt_p, false);
-		tcpip_bindports((uint16_t)opt_p, true);
-		while (tcpip_connect((uint16_t)opt_p, false) < peers.size())
-			sleep(1);
-		while (tcpip_connect((uint16_t)opt_p, true) < peers.size())
-			sleep(1);
-		tcpip_accept();
-		tcpip_fork();
-		ret = tcpip_io();
-		tcpip_close();
-		tcpip_done();
-		// release cache
-		tmcg_mpz_ssrandomm_cache_done(cache, cache_mod, cache_avail);
-		// finish
-		return ret;
+		// start interactive variant, if built-in TCP/IP requested
+		ret = run_tcpip(peers.size(), opt_p, hostname, port);
 	}
 	else if (opt_y)
 	{
-		// run as replacement for GnuPG et al. (yet-another-openpgp-tool)
-		run_instance(0, time(NULL), opt_e, 0);
-		// release cache
-		tmcg_mpz_ssrandomm_cache_done(cache, cache_mod, cache_avail);
-		return ret;
+		// start a single instance as replacement for GnuPG et al.
+		fork_instance(0);
+		ret = wait_instance(0, opt_verbose, pid);
 	}
-
-#ifdef GNUNET
-	static const struct GNUNET_GETOPT_CommandLineOption myoptions[] = {
-		GNUNET_GETOPT_option_uint('e',
-			"expiration",
-			"INTEGER",
-			"expiration time of generated keys in seconds",
-			&gnunet_opt_keyexptime
-		),
-		GNUNET_GETOPT_option_string('g',
-			"group",
-			"STRING",
-			"common reference string that defines the underlying DDH-hard group",
-			&gnunet_opt_crs
-		),
-		GNUNET_GETOPT_option_string('H',
-			"hostname",
-			"STRING",
-			"hostname (e.g. onion address) of this peer within PEERS",
-			&gnunet_opt_hostname
-		),
-		GNUNET_GETOPT_option_string('p',
-			"port",
-			"STRING",
-			"GNUnet CADET port to listen/connect",
-			&gnunet_opt_port
-		),
-		GNUNET_GETOPT_option_string('P',
-			"passwords",
-			"STRING",
-			"exchanged passwords to protect private and broadcast channels",
-			&gnunet_opt_passwords
-		),
-		GNUNET_GETOPT_option_uint('s',
-			"s-resilience",
-			"INTEGER",
-			"resilience of threshold DSS protocol (signature scheme)",
-			&gnunet_opt_s_resilience
-		),
-		GNUNET_GETOPT_option_uint('t',
-			"t-resilience",
-			"INTEGER",
-			"resilience of DKG protocol (threshold decryption)",
-			&gnunet_opt_t_resilience
-		),
-		GNUNET_GETOPT_option_flag('T',
-			"timestamping",
-			"state that the generated key is used for timestamping",
-			&gnunet_opt_timestamping
-		),
-		GNUNET_GETOPT_option_flag('V',
-			"verbose",
-			"turn on verbose output",
-			&gnunet_opt_verbose
-		),
-		GNUNET_GETOPT_option_uint('w',
-			"wait",
-			"INTEGER",
-			"minutes to wait until start of key generation protocol",
-			&gnunet_opt_wait
-		),
-		GNUNET_GETOPT_option_uint('W',
-			"aiou-timeout",
-			"INTEGER",
-			"timeout for point-to-point messages in minutes",
-			&gnunet_opt_W
-		),
-		GNUNET_GETOPT_option_uint('x',
-			"x-tests",
-			NULL,
-			"number of exchange tests",
-			&gnunet_opt_xtests
-		),
-		GNUNET_GETOPT_option_flag('y',
-			"yaot",
-			"yet another OpenPGP tool",
-			&gnunet_opt_y
-		),
-		GNUNET_GETOPT_OPTION_END
-	};
-	ret = GNUNET_PROGRAM_run(argc, argv, usage, about, myoptions, &gnunet_run,
-		argv[0]);
-//	GNUNET_free((void *) argv);
-	// release cache
-	tmcg_mpz_ssrandomm_cache_done(cache, cache_mod, cache_avail);
-	// finish
-	if (ret == GNUNET_OK)
-		return 0;
 	else
-		return -1;
+	{
+#ifdef GNUNET
+		static const struct GNUNET_GETOPT_CommandLineOption myoptions[] = {
+			GNUNET_GETOPT_option_uint('e',
+				"expiration",
+				"INTEGER",
+				"expiration time of generated keys in seconds",
+				&gnunet_opt_keyexptime
+			),
+			GNUNET_GETOPT_option_string('g',
+				"group",
+				"STRING",
+				"common reference string that defines the underlying DDH-hard group",
+				&gnunet_opt_crs
+			),
+			GNUNET_GETOPT_option_string('H',
+				"hostname",
+				"STRING",
+				"hostname (e.g. onion address) of this peer within PEERS",
+				&gnunet_opt_hostname
+			),
+			GNUNET_GETOPT_option_flag('N',
+				"no-passphrase",
+				"disable private key protection",
+				&gnunet_opt_nopassphrase
+			),
+			GNUNET_GETOPT_option_string('p',
+				"port",
+				"STRING",
+				"GNUnet CADET port to listen/connect",
+				&gnunet_opt_port
+			),
+			GNUNET_GETOPT_option_string('P',
+				"passwords",
+				"STRING",
+				"exchanged passwords to protect private and broadcast channels",
+				&gnunet_opt_passwords
+			),
+			GNUNET_GETOPT_option_uint('s',
+				"s-resilience",
+				"INTEGER",
+				"resilience of threshold DSS protocol (signature scheme)",
+				&gnunet_opt_s_resilience
+			),
+			GNUNET_GETOPT_option_uint('t',
+				"t-resilience",
+				"INTEGER",
+				"resilience of DKG protocol (threshold decryption)",
+				&gnunet_opt_t_resilience
+			),
+			GNUNET_GETOPT_option_flag('T',
+				"timestamping",
+				"state that the generated key is used for timestamping",
+				&gnunet_opt_timestamping
+			),
+			GNUNET_GETOPT_option_string('u',
+				"uid",
+				"STRING",
+				"user ID of the generated key",
+				&gnunet_opt_u
+			),
+			GNUNET_GETOPT_option_flag('V',
+				"verbose",
+				"turn on verbose output",
+				&gnunet_opt_verbose
+			),
+			GNUNET_GETOPT_option_uint('w',
+				"wait",
+				"INTEGER",
+				"minutes to wait until start of key generation protocol",
+				&gnunet_opt_wait
+			),
+			GNUNET_GETOPT_option_uint('W',
+				"aiou-timeout",
+				"INTEGER",
+				"timeout for point-to-point messages in minutes",
+				&gnunet_opt_W
+			),
+			GNUNET_GETOPT_option_uint('x',
+				"x-tests",
+				NULL,
+				"number of exchange tests",
+				&gnunet_opt_xtests
+			),
+			GNUNET_GETOPT_option_flag('y',
+				"yaot",
+				"yet another OpenPGP tool",
+				&gnunet_opt_y
+			),
+			GNUNET_GETOPT_OPTION_END
+		};
+		ret = GNUNET_PROGRAM_run(argc, argv, usage, about, myoptions,
+			&gnunet_run, argv[0]);
+		if (ret != GNUNET_OK)
+			ret = -1;
 #else
-	std::cerr << "WARNING: GNUnet CADET is required for the message exchange" <<
-		" of this program" << std::endl;
+		ret = run_localtest(peers.size(), opt_verbose, pid, pipefd,
+			broadcast_pipefd, &fork_instance);
 #endif
-
-	std::cerr << "INFO: running local test with " << peers.size() <<
-		" participants" << std::endl;
-	std::cerr << "WARNING: due to cache issues the generated shares will be" <<
-		" identical, don't use them!" << std::endl;
-	// open pipes
-	for (size_t i = 0; i < peers.size(); i++)
-	{
-		for (size_t j = 0; j < peers.size(); j++)
-		{
-			if (pipe(pipefd[i][j]) < 0)
-				perror("ERROR: dkg-generate (pipe)");
-			if (pipe(broadcast_pipefd[i][j]) < 0)
-				perror("ERROR: dkg-generate (pipe)");
-		}
 	}
-	
-	// start childs
-	for (size_t i = 0; i < peers.size(); i++)
-		fork_instance(i);
-
-	// sleep for five seconds
-	sleep(5);
-	
-	// wait for childs and close pipes
-	for (size_t i = 0; i < peers.size(); i++)
-	{
-		int wstatus = 0;
-		if (opt_verbose)
-			std::cerr << "INFO: waitpid(" << pid[i] << ")" << std::endl;
-		if (waitpid(pid[i], &wstatus, 0) != pid[i])
-			perror("ERROR: dkg-generate (waitpid)");
-		if (!WIFEXITED(wstatus))
-		{
-			std::cerr << "ERROR: protocol instance ";
-			if (WIFSIGNALED(wstatus))
-			{
-				std::cerr << pid[i] << " terminated by signal " <<
-					WTERMSIG(wstatus) << std::endl;
-			}
-			if (WCOREDUMP(wstatus))
-				std::cerr << pid[i] << " dumped core" << std::endl;
-			ret = -1; // fatal error
-		}
-		else if (WIFEXITED(wstatus))
-		{
-			if (opt_verbose)
-			{
-				std::cerr << "INFO: protocol instance " << pid[i] <<
-					" terminated with exit status " << WEXITSTATUS(wstatus) <<
-					std::endl;
-			}
-			if (WEXITSTATUS(wstatus))
-				ret = -2; // error
-		}
-		for (size_t j = 0; j < peers.size(); j++)
-		{
-			if ((close(pipefd[i][j][0]) < 0) || (close(pipefd[i][j][1]) < 0))
-				perror("ERROR: dkg-generate (close)");
-			if ((close(broadcast_pipefd[i][j][0]) < 0) ||
-				(close(broadcast_pipefd[i][j][1]) < 0))
-			{
-				perror("ERROR: dkg-generate (close)");
-			}
-		}
-	}
-	
 	// release cache
 	tmcg_mpz_ssrandomm_cache_done(cache, cache_mod, cache_avail);
 	// finish
+	if (should_unlock)
+		unlock_memory();
 	return ret;
 }
 
