@@ -1,7 +1,7 @@
 /*******************************************************************************
    This file is part of Distributed Privacy Guard (DKGPG).
 
- Copyright (C) 2017, 2018  Heiko Stamer <HeikoStamer@gmx.net>
+ Copyright (C) 2017, 2018, 2019  Heiko Stamer <HeikoStamer@gmx.net>
 
    DKGPG is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -24,10 +24,12 @@
 #endif
 #ifdef DKGPG_TESTSUITE
 	#undef GNUNET
-#else
+#endif
 #ifdef DKGPG_TESTSUITE_Y
 	#undef GNUNET
 #endif
+#ifdef DKGPG_TESTSUITE_TS
+	#undef GNUNET
 #endif
 #include "dkg-gnunet-common.hh"
 
@@ -83,6 +85,8 @@ std::map<size_t, struct GNUNET_CADET_Channel*> 	pipe2channel_in;
 void gnunet_hello_callback
 	(void *cls, const struct GNUNET_MessageHeader *hello)
 {
+	if (cls == NULL)
+		cls = NULL; // dummy code to supress compiler warning
 	if (hello == NULL)
 	{
 		GNUNET_log(GNUNET_ERROR_TYPE_ERROR, "No hello message in callback\n");
@@ -105,6 +109,8 @@ void gnunet_hello_callback
 int check_gnunet_data_callback
 	(void *cls, const struct GNUNET_MessageHeader *message)
 {
+	if ((cls == NULL) && (message == NULL))
+		cls = NULL; // dummy code to supress compiler warning
 	return GNUNET_OK;
 }
 
@@ -194,6 +200,8 @@ void handle_gnunet_data_callback
 void gnunet_pipe_ready
 	(void *cls)
 {
+	if (cls == NULL)
+		cls = NULL; // dummy code to supress compiler warning
 	pt = NULL;
 	for (size_t i = 0; i < peers.size(); i++)
 	{
@@ -238,6 +246,8 @@ void gnunet_pipe_ready
 void gnunet_broadcast_pipe_ready
 	(void *cls)
 {
+	if (cls == NULL)
+		cls = NULL; // dummy code to supress compiler warning
 	pt_broadcast = NULL;
 	for (size_t i = 0; i < peers.size(); i++)
 	{
@@ -282,6 +292,8 @@ void gnunet_broadcast_pipe_ready
 void gnunet_channel_ended
 	(void *cls, const struct GNUNET_CADET_Channel *channel)
 {
+	if (cls == NULL)
+		cls = NULL; // dummy code to supress compiler warning
 	// deregister the ended channel	
 	for (size_t i = 0; i < peers.size(); i++)
 	{
@@ -309,6 +321,8 @@ void* gnunet_channel_incoming
 	(void *cls, struct GNUNET_CADET_Channel *channel,
 	 const struct GNUNET_PeerIdentity *initiator)
 {
+	if (cls == NULL)
+		cls = NULL; // dummy code to supress compiler warning
 	if (gnunet_opt_verbose)
 		std::cerr << "INFO: incoming channel from " <<
 			GNUNET_i2s_full(initiator) << std::endl;
@@ -334,6 +348,8 @@ void* gnunet_channel_incoming
 void gnunet_shutdown_task
 	(void *cls)
 {
+	if (cls == NULL)
+		cls = NULL; // dummy code to supress compiler warning
 	GNUNET_log(GNUNET_ERROR_TYPE_DEBUG, "Shutdown\n");
 	// cancel scheduled tasks
 	if (pt_broadcast != NULL)
@@ -372,14 +388,14 @@ void gnunet_shutdown_task
 		job = NULL;
 	}
 	// release buffered messages
-	while (send_queue.size())
+	while (!send_queue.empty())
 	{
 		DKG_BufferListEntry ble = send_queue.front();
 		DKG_Buffer qbuf = ble.second;
 		delete [] qbuf.second;
 		send_queue.pop_front();
 	}
-	while (send_queue_broadcast.size())
+	while (!send_queue_broadcast.empty())
 	{
 		DKG_BufferListEntry ble = send_queue_broadcast.front();
 		DKG_Buffer qbuf = ble.second;
@@ -454,10 +470,12 @@ void gnunet_shutdown_task
 void gnunet_io
 	(void *cls)
 {
+	if (cls == NULL)
+		cls = NULL; // dummy code to supress compiler warning
 	io = NULL;
 
 	// send messages to peers
-	if (send_queue.size())
+	if (!send_queue.empty())
 	{
 		DKG_BufferListEntry ble = send_queue.front();
 		if (pipe2channel_in.count(ble.first)) // have input channel to this peer?
@@ -481,7 +499,7 @@ void gnunet_io
 	}
 
 	// send broadcast messages to peers
-	if (send_queue_broadcast.size())
+	if (!send_queue_broadcast.empty())
 	{
 		DKG_BufferListEntry ble = send_queue_broadcast.front();
 		if (pipe2channel_in.count(ble.first)) // have input channel to this peer?
@@ -505,7 +523,7 @@ void gnunet_io
 	}
 
 	// reschedule I/O task, if further messages available
-	if (send_queue.size() || send_queue_broadcast.size())
+	if (!send_queue.empty() || !send_queue_broadcast.empty())
 		io = GNUNET_SCHEDULER_add_now(&gnunet_io, NULL);
 
 	// schedule tasks for reading the input pipes 
@@ -525,6 +543,8 @@ void gnunet_io
 void gnunet_connect
 	(void *cls)
 {
+	if (cls == NULL)
+		cls = NULL; // dummy code to supress compiler warning
 	ct = NULL;
 
 	for (size_t i = 0; i < peers.size(); i++)
@@ -593,6 +613,8 @@ void gnunet_connect
 void gnunet_statistics
 	(void *cls)
 {
+	if (cls == NULL)
+		cls = NULL; // dummy code to supress compiler warning
 	st = NULL;
 
 	if ((gnunet_opt_verbose) && (opt_verbose > 1))
@@ -641,6 +663,8 @@ void gnunet_statistics
 void gnunet_fork
 	(void *cls)
 {
+	if (cls == NULL)
+		cls = NULL; // dummy code to supress compiler warning
 	ft = NULL;
 
 	if (pipe2channel_in.size() == (peers.size() - 1))
@@ -661,6 +685,8 @@ void gnunet_fork
 void gnunet_init
 	(void *cls)
 {
+	if (cls == NULL)
+		cls = NULL; // dummy code to supress compiler warning
 	job = NULL;
 
 	// wait until we got our own peer identity from TRANSPORT
@@ -709,6 +735,9 @@ void gnunet_init
 void gnunet_run(void *cls, char *const *args, const char *cfgfile,
 	const struct GNUNET_CONFIGURATION_Handle *cfg)
 {
+	if ((args == NULL) && (cfgfile == NULL)) 
+		args = NULL; // dummy code to supress compiler warning
+
 	// initialize peer2pipe and pipe2peer mapping
 	for (size_t i = 0; i < peers.size(); i++)
 	{
