@@ -1,7 +1,7 @@
 /*******************************************************************************
    This file is part of Distributed Privacy Guard (DKGPG).
 
- Copyright (C) 2017, 2018  Heiko Stamer <HeikoStamer@gmx.net>
+ Copyright (C) 2017, 2018, 2019  Heiko Stamer <HeikoStamer@gmx.net>
 
    DKGPG is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -104,7 +104,7 @@ void run_instance
 	std::string armored_pubring;
 	if (opt_k)
 	{
-		if (!read_key_file(kfilename, armored_pubring))
+		if (!read_key_file(opt_k, armored_pubring))
 			exit(-1);
 	}
 
@@ -159,17 +159,16 @@ void run_instance
 		delete ring;
 		exit(-1);
 	}
+	delete ring;
 	if (!prv->pub->valid || ((opt_y == NULL) && prv->Weak(opt_verbose)))
 	{
 		std::cerr << "ERROR: primary key is invalid or weak" << std::endl;
-		delete ring;
 		delete prv;
 		exit(-1);
 	}
 	if ((prv->pkalgo != TMCG_OPENPGP_PKALGO_EXPERIMENTAL7) && (opt_y == NULL))
 	{
 		std::cerr << "ERROR: primary key is not a tDSS/DSA key" << std::endl;
-		delete ring;
 		delete prv;
 		exit(-1);
 	}
@@ -187,7 +186,6 @@ void run_instance
 		if (!init_tDSS(prv, opt_verbose, dss))
 		{
 			delete dss;
-			delete ring;
 			delete prv;
 			exit(-1);
 		}
@@ -197,7 +195,6 @@ void run_instance
 			std::cerr << "ERROR: creating 1-to-1 CAPL mapping failed" <<
 				std::endl;
 			delete dss;
-			delete ring;
 			delete prv;
 			exit(-1);
 		}
@@ -212,11 +209,10 @@ void run_instance
 				std::string pwd;
 				if (!TMCG_ParseHelper::gs(passwords, '/', pwd))
 				{
-					std::cerr << "ERROR: S_" << whoami << ": " <<
-						"cannot read" << " password for protecting channel" <<
-						" to S_" << i << std::endl;
+					std::cerr << "ERROR: p_" << whoami << ": " <<
+						"cannot read password for protecting channel to p_" <<
+						i << std::endl;
 					delete dss;
-					delete ring;
 					delete prv;
 					exit(-1);
 				}
@@ -224,11 +220,10 @@ void run_instance
 				if (((i + 1) < peers.size()) &&
 					!TMCG_ParseHelper::nx(passwords, '/'))
 				{
-					std::cerr << "ERROR: S_" << whoami << ": " << "cannot" <<
-						" skip to next password for protecting channel to S_" <<
+					std::cerr << "ERROR: p_" << whoami << ": " << "cannot" <<
+						" skip to next password for protecting channel to p_" <<
 						(i + 1) << std::endl;
 					delete dss;
-					delete ring;
 					delete prv;
 					exit(-1);
 				}
@@ -236,7 +231,7 @@ void run_instance
 			else
 			{
 				// simple key -- we assume that GNUnet provides secure channels
-				key << "dkg-sign::S_" << (i + whoami);
+				key << "dkg-sign::p_" << (i + whoami);
 			}
 			uP_in.push_back(pipefd[i][whoami][0]);
 			uP_out.push_back(pipefd[whoami][i][1]);
@@ -263,82 +258,17 @@ void run_instance
 				aiou2, aiounicast::aio_scheduler_roundrobin, (opt_W * 60));
 		rbc->setID(myID);
 		// perform a simple exchange test with debug output
-		for (size_t i = 0; i < num_xtests; i++)
-		{
-			mpz_t xtest;
-			mpz_init_set_ui(xtest, i);
-			std::cerr << "INFO: S_" << whoami << ": xtest = " << xtest <<
-				" <-> ";
-			rbc->Broadcast(xtest);
-			for (size_t ii = 0; ii < peers.size(); ii++)
-			{
-				if (!rbc->DeliverFrom(xtest, ii))
-					std::cerr << "<X> ";
-				else
-					std::cerr << xtest << " ";
-			}
-			std::cerr << std::endl;
-			mpz_clear(xtest);
-		}
+		xtest(num_xtests, whoami, peers.size(), rbc);
 		// participants must agree on a common signature creation time (OpenPGP)
-		if (opt_verbose)
-			std::cerr << "INFO: agree on a signature creation time for" <<
-				" OpenPGP" << std::endl;
-		std::vector<time_t> tvs;
-		mpz_t mtv;
-		mpz_init_set_ui(mtv, sigtime);
-		rbc->Broadcast(mtv);
-		tvs.push_back(sigtime);
-		for (size_t i = 0; i < peers.size(); i++)
-		{
-			if (i != whoami)
-			{
-				if (rbc->DeliverFrom(mtv, i))
-				{
-					time_t utv;
-					utv = (time_t)mpz_get_ui(mtv);
-					tvs.push_back(utv);
-				}
-				else
-				{
-					std::cerr << "WARNING: S_" << whoami << ": no signature" <<
-						" creation time stamp received from S_" << i <<
-						std::endl;
-				}
-			}
-		}
-		mpz_clear(mtv);
-		std::sort(tvs.begin(), tvs.end());
-		if (tvs.size() < (peers.size() - T_RBC))
-		{
-			std::cerr << "ERROR: S_" << whoami << ": not enough timestamps" <<
-				" received" << std::endl;
-			delete rbc, delete aiou, delete aiou2;
-			delete dss;
-			delete ring;
-			delete prv;
-			exit(-1);
-		}
-		// use a median value as some kind of gentle agreement
-		csigtime = tvs[tvs.size()/2];
-		if (opt_verbose)
-			std::cerr << "INFO: S_" << whoami << ": canonicalized signature" <<
-				" creation time = " << csigtime << std::endl;
+		csigtime = agree_time(sigtime, whoami, peers.size(), opt_verbose, rbc);
 		// select hash algorithm for OpenPGP based on |q| (size in bit)
-		if (mpz_sizeinbase(dss->q, 2L) == 256)
-			hashalgo = TMCG_OPENPGP_HASHALGO_SHA256; // SHA256 (alg 8)
-		else if (mpz_sizeinbase(dss->q, 2L) == 384)
-			hashalgo = TMCG_OPENPGP_HASHALGO_SHA384; // SHA384 (alg 9)
-		else if (mpz_sizeinbase(dss->q, 2L) == 512)
-			hashalgo = TMCG_OPENPGP_HASHALGO_SHA512; // SHA512 (alg 10)
-		else
+		if (!select_hashalgo(dss, hashalgo))
 		{
-			std::cerr << "ERROR: S_" << whoami << ": selecting hash" <<
+			std::cerr << "ERROR: p_" << whoami << ": selecting hash" <<
 				" algorithm failed for |q| = " << mpz_sizeinbase(dss->q, 2L) <<
 				std::endl;
 			delete rbc, delete aiou, delete aiou2;
 			delete dss;
-			delete ring;
 			delete prv;
 			exit(-1);
 		}
@@ -375,7 +305,7 @@ void run_instance
 		if (!CallasDonnerhackeFinneyShawThayerRFC4880::
 			TextDocumentHash(opt_ifilename, trailer, hashalgo, hash, left))
 		{
-			std::cerr << "ERROR: S_" << whoami << ": TextDocumentHash()" <<
+			std::cerr << "ERROR: p_" << whoami << ": TextDocumentHash()" <<
 				" failed; cannot process input file \"" << opt_ifilename <<
 				"\"" << std::endl;
 			if (opt_y == NULL)
@@ -383,7 +313,6 @@ void run_instance
 				delete rbc, delete aiou, delete aiou2;
 				delete dss;
 			}
-			delete ring;
 			delete prv;
 			exit(-1);
 		}
@@ -407,7 +336,7 @@ void run_instance
 		if (!CallasDonnerhackeFinneyShawThayerRFC4880::
 			BinaryDocumentHash(opt_ifilename, trailer, hashalgo, hash, left))
 		{
-			std::cerr << "ERROR: S_" << whoami << ": BinaryDocumentHash()" <<
+			std::cerr << "ERROR: p_" << whoami << ": BinaryDocumentHash()" <<
 				" failed; cannot process input file \"" << opt_ifilename <<
 				"\"" << std::endl;
 			if (opt_y == NULL)
@@ -415,183 +344,32 @@ void run_instance
 				delete rbc, delete aiou, delete aiou2;
 				delete dss;
 			}
-			delete ring;
 			delete prv;
 			exit(-1);
 		}
 	}
 
-	// prepare the hash value
-	tmcg_openpgp_byte_t buffer[1024];
-	size_t buflen = 0;
-	memset(buffer, 0, sizeof(buffer));
-	if (opt_verbose > 1)
-		std::cerr << std::hex << "INFO: hash = ";
-	for (size_t i = 0; i < hash.size(); i++, buflen++)
-	{
-		if (i < sizeof(buffer))
-			buffer[i] = hash[i];
-		if (opt_verbose > 1)
-			std::cerr << (int)hash[i] << " ";
-	}
-	if (opt_verbose > 1)
-		std::cerr << std::dec << std::endl;
-
 	// sign the hash value
-	gcry_error_t ret;
-	gcry_mpi_t r, s;
-	r = gcry_mpi_new(2048);
-	s = gcry_mpi_new(2048);
-	if (opt_y == NULL)
-	{
-		gcry_mpi_t h;
-		ret = gcry_mpi_scan(&h, GCRYMPI_FMT_USG, buffer, buflen, NULL);
-		if (ret)
-		{
-			std::cerr << "ERROR: S_" << whoami << ": gcry_mpi_scan() failed" <<
-				" for h" << std::endl;
-			delete rbc, delete aiou, delete aiou2;
-			delete dss;
-			delete ring;
-			delete prv;
-			exit(-1);
-		}
-		if (opt_verbose > 1)
-			std::cerr << "INFO: S_" << whoami << ": h = " << h << std::endl;
-		mpz_t dsa_m, dsa_r, dsa_s;
-		mpz_init(dsa_m), mpz_init(dsa_r), mpz_init(dsa_s);
-		if (!tmcg_mpz_set_gcry_mpi(h, dsa_m))
-		{
-			std::cerr << "ERROR: S_" << whoami << ": tmcg_mpz_set_gcry_mpi()" <<
-				" failed for dsa_m" << std::endl;
-			gcry_mpi_release(r), gcry_mpi_release(s), gcry_mpi_release(h);
-			mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-			delete rbc, delete aiou, delete aiou2;
-			delete dss;
-			delete ring;
-			delete prv;
-			exit(-1);
-		}
-		gcry_mpi_release(h);
-		std::stringstream err_log_sign;
-		if (opt_verbose)
-			std::cerr << "INFO: S_" << whoami << ": dss.Sign()" << std::endl;
-		if (!dss->Sign(peers.size(), whoami, dsa_m, dsa_r, dsa_s,
-			prv->tdss_idx2dkg, prv->tdss_dkg2idx, aiou, rbc, err_log_sign))
-		{
-			std::cerr << "ERROR: S_" << whoami << ": " <<
-				"tDSS Sign() failed" << std::endl;
-			std::cerr << "ERROR: S_" << whoami << ": log follows " <<
-				std::endl << err_log_sign.str();
-			gcry_mpi_release(r), gcry_mpi_release(s);
-			mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-			delete rbc, delete aiou, delete aiou2;
-			delete dss;
-			delete ring;
-			delete prv;
-			exit(-1);
-		}
-		if (opt_verbose > 1)
-			std::cerr << "INFO: S_" << whoami << ": log follows " <<
-				std::endl << err_log_sign.str();
-		if (!tmcg_mpz_get_gcry_mpi(r, dsa_r))
-		{
-			std::cerr << "ERROR: S_" << whoami << ": tmcg_mpz_get_gcry_mpi()" <<
-				" failed for dsa_r" << std::endl;
-			gcry_mpi_release(r), gcry_mpi_release(s);
-			mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-			delete rbc, delete aiou, delete aiou2;
-			delete dss;
-			delete ring;
-			delete prv;
-			exit(-1);
-		}
-		if (!tmcg_mpz_get_gcry_mpi(s, dsa_s))
-		{
-			std::cerr << "ERROR: S_" << whoami << ": tmcg_mpz_get_gcry_mpi()" <<
-				" failed for dsa_s" << std::endl;
-			gcry_mpi_release(r), gcry_mpi_release(s);
-			mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-			delete rbc, delete aiou, delete aiou2;
-			delete dss;
-			delete ring;
-			delete prv;
-			exit(-1);
-		}
-		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-	}
-	else
-	{
-		switch (prv->pkalgo)
-		{
-			case TMCG_OPENPGP_PKALGO_RSA:
-			case TMCG_OPENPGP_PKALGO_RSA_SIGN_ONLY:
-				ret = CallasDonnerhackeFinneyShawThayerRFC4880::
-					AsymmetricSignRSA(hash, prv->private_key, hashalgo, s);
-				break;
-			case TMCG_OPENPGP_PKALGO_DSA:
-				ret = CallasDonnerhackeFinneyShawThayerRFC4880::
-					AsymmetricSignDSA(hash, prv->private_key, r, s);
-				break;
-			case TMCG_OPENPGP_PKALGO_ECDSA:
-				ret = CallasDonnerhackeFinneyShawThayerRFC4880::
-					AsymmetricSignECDSA(hash, prv->private_key, r, s);
-				break;
-			case TMCG_OPENPGP_PKALGO_EDDSA:
-				ret = CallasDonnerhackeFinneyShawThayerRFC4880::
-					AsymmetricSignEdDSA(hash, prv->private_key, r, s);
-				break;
-			default:
-				std::cerr << "ERROR: public-key algorithm " <<
-					(int)prv->pkalgo << " not supported" << std::endl;
-				gcry_mpi_release(r), gcry_mpi_release(s);
-				delete ring;
-				delete prv;
-				exit(-1);
-		}
-		if (ret)
-		{
-			std::cerr << "ERROR: signing of hash value failed " <<
-				"(rc = " << gcry_err_code(ret) << ", str = " <<
-				gcry_strerror(ret) << ")" << std::endl;
-			gcry_mpi_release(r), gcry_mpi_release(s);
-			delete ring;
-			delete prv;
-			exit(-1);
-		}
-	}
 	tmcg_openpgp_octets_t sig;
-	switch (prv->pkalgo)
+	if (!sign_hash(hash, trailer, left, whoami, peers.size(), prv, hashalgo,
+		sig, opt_verbose, opt_y, dss, aiou, rbc))
 	{
-		case TMCG_OPENPGP_PKALGO_RSA:
-		case TMCG_OPENPGP_PKALGO_RSA_SIGN_ONLY:
-			CallasDonnerhackeFinneyShawThayerRFC4880::
-				PacketSigEncode(trailer, left, s, sig);
-			break;
-		case TMCG_OPENPGP_PKALGO_DSA:
-		case TMCG_OPENPGP_PKALGO_ECDSA:
-		case TMCG_OPENPGP_PKALGO_EDDSA:
-		case TMCG_OPENPGP_PKALGO_EXPERIMENTAL7:
-			CallasDonnerhackeFinneyShawThayerRFC4880::
-				PacketSigEncode(trailer, left, r, s, sig);
-			break;
-		default:
-			std::cerr << "ERROR: public-key algorithm " << (int)prv->pkalgo <<
-				" not supported" << std::endl;
-			gcry_mpi_release(r), gcry_mpi_release(s);
-			delete ring;
-			delete prv;
-			exit(-1);
+		if (opt_y == NULL)
+		{
+			delete rbc, delete aiou, delete aiou2;
+			delete dss;
+		}
+		delete prv;
+		exit(-1);
 	}
-	gcry_mpi_release(r), gcry_mpi_release(s);
 
 	// release allocated ressources
 	if (opt_y == NULL)
 	{
 		// at the end: deliver some more rounds for still waiting parties
-		time_t synctime = aiounicast::aio_timeout_long;
+		time_t synctime = (opt_W * 6);
 		if (opt_verbose)
-			std::cerr << "INFO: S_" << whoami << ": waiting approximately " <<
+			std::cerr << "INFO: p_" << whoami << ": waiting approximately " <<
 				(synctime * (T_RBC + 1)) << " seconds for stalled parties" <<
 				std::endl;
 		rbc->Sync(synctime);
@@ -600,14 +378,14 @@ void run_instance
 		// release handles (unicast channel)
 		if (opt_verbose)
 		{
-			std::cerr << "INFO: S_" << whoami << ": unicast channels";
+			std::cerr << "INFO: p_" << whoami << ": unicast channels";
 			aiou->PrintStatistics(std::cerr);
 			std::cerr << std::endl;
 		}
 		// release handles (broadcast channel)
 		if (opt_verbose)
 		{
-			std::cerr << "INFO: S_" << whoami << ": broadcast channel";
+			std::cerr << "INFO: p_" << whoami << ": broadcast channel";
 			aiou2->PrintStatistics(std::cerr);
 			std::cerr << std::endl;
 		}
@@ -616,7 +394,6 @@ void run_instance
 		// release threshold signature scheme
 		delete dss;
 	}
-	delete ring;
 	delete prv;
 
 	// output the result
@@ -635,7 +412,7 @@ void run_instance
 		if (!CallasDonnerhackeFinneyShawThayerRFC4880::
 			DashEscapeFile(opt_ifilename, ct_body))
 		{
-			std::cerr << "ERROR: S_" << whoami << ": DashEscapeFile()" <<
+			std::cerr << "ERROR: p_" << whoami << ": DashEscapeFile()" <<
 				" failed; cannot process input file \"" << opt_ifilename <<
 				"\"" << std::endl;
 			exit(-1);
@@ -679,7 +456,7 @@ void fork_instance
 	{
 		if (pid[whoami] == 0)
 		{
-			/* BEGIN child code: participant S_i */
+			/* BEGIN child code: participant p_i */
 			time_t sigtime = time(NULL);
 #ifdef GNUNET
 			run_instance(whoami, sigtime, gnunet_opt_sigexptime,
@@ -688,9 +465,9 @@ void fork_instance
 			run_instance(whoami, sigtime, opt_e, 0);
 #endif
 			if (opt_verbose)
-				std::cerr << "INFO: S_" << whoami << ": exit(0)" << std::endl;
+				std::cerr << "INFO: p_" << whoami << ": exit(0)" << std::endl;
 			exit(0);
-			/* END child code: participant S_i */
+			/* END child code: participant p_i */
 		}
 		else
 		{
@@ -719,7 +496,7 @@ int main
 		GNUNET_GETOPT_option_help(about),
 		GNUNET_GETOPT_option_uint('e',
 			"expiration",
-			"TIME",
+			"INTEGER",
 			"expiration time of generated signature in seconds",
 			&gnunet_opt_sigexptime
 		),
@@ -785,13 +562,13 @@ int main
 		),
 		GNUNET_GETOPT_option_uint('w',
 			"wait",
-			"TIME",
+			"INTEGER",
 			"minutes to wait until start of signing protocol",
 			&gnunet_opt_wait
 		),
 		GNUNET_GETOPT_option_uint('W',
 			"aiou-timeout",
-			"TIME",
+			"INTEGER",
 			"timeout for point-to-point messages in minutes",
 			&gnunet_opt_W
 		),
@@ -841,7 +618,7 @@ int main
 	if (gnunet_opt_URI != NULL)
 		URI = gnunet_opt_URI; // get policy URI from GNUnet options
 	if (gnunet_opt_y != NULL)
-		opt_y = gnunet_opt_y; // get yaot filename from GNUnet options
+		opt_y = gnunet_opt_y;
 #endif
 
 	// create peer list from remaining arguments
@@ -933,7 +710,7 @@ int main
 				std::cout << "  -C, --clear    apply cleartext signature" <<
 					" framework (cf. RFC 4880)" << std::endl;
 				std::cout << "  -h, --help     print this help" << std::endl;
-				std::cout << "  -e TIME        expiration time of generated" <<
+				std::cout << "  -e INTEGER     expiration time of generated" <<
 					" signature in seconds" << std::endl;
 				std::cout << "  -E, --echo     enable terminal echo when" <<
 					" reading passphrase" << std::endl;
@@ -957,7 +734,7 @@ int main
 					std::endl;
 				std::cout << "  -V, --verbose  turn on verbose output" <<
 					std::endl;
-				std::cout << "  -W TIME        timeout for point-to-point" <<
+				std::cout << "  -W INTEGER     timeout for point-to-point" <<
 					" messages in minutes" << std::endl;
 				std::cout << "  -y FILENAME    yet another OpenPGP tool with" <<
 					" private key in FILENAME" << std::endl;
@@ -1042,12 +819,7 @@ int main
 			usage << std::endl;
 		return -1;
 	}
-
-	// canonicalize peer list
-	std::sort(peers.begin(), peers.end());
-	std::vector<std::string>::iterator it =
-		std::unique(peers.begin(), peers.end());
-	peers.resize(std::distance(peers.begin(), it));
+	canonicalize(peers);
 	if (((peers.size() < 3)  || (peers.size() > DKGPG_MAX_N)) && (opt_y == NULL))
 	{
 		std::cerr << "ERROR: too few or too many peers given" << std::endl;
@@ -1061,7 +833,7 @@ int main
 	}
 
 	// lock memory
-	bool force_secmem = false;
+	bool force_secmem = false, should_unlock = false;
 	if (!lock_memory())
 	{
 		std::cerr << "WARNING: locking memory failed; CAP_IPC_LOCK required" <<
@@ -1069,219 +841,144 @@ int main
 		// at least try to use libgcrypt's secure memory
 		force_secmem = true;
 	}
+	else
+		should_unlock = true;
 
 	// initialize LibTMCG
 	if (!init_libTMCG(force_secmem))
 	{
 		std::cerr << "ERROR: initialization of LibTMCG failed" << std::endl;
+		if (should_unlock)
+			unlock_memory();
 		return -1;
 	}
 	if (opt_verbose)
 		std::cerr << "INFO: using LibTMCG version " << version_libTMCG() <<
 			std::endl;
 	
-	// initialize return code
+	// initialize return code and do the main work
 	int ret = 0;
-	// create underlying point-to-point channels, if built-in TCP/IP requested
 	if ((opt_hostname != NULL) && (opt_y == NULL))
 	{
-		if (port.length())
-			opt_p = strtoul(port.c_str(), NULL, 10); // start port from options
-		if ((opt_p < 1) || (opt_p > 65535))
-		{
-			std::cerr << "ERROR: no valid TCP start port given" << std::endl;
-			return -1;
-		}
-		tcpip_init(hostname);
-		tcpip_bindports((uint16_t)opt_p, false);
-		tcpip_bindports((uint16_t)opt_p, true);
-		while (tcpip_connect((uint16_t)opt_p, false) < peers.size())
-			sleep(1);
-		while (tcpip_connect((uint16_t)opt_p, true) < peers.size())
-			sleep(1);
-		tcpip_accept();
-		tcpip_fork();
-		ret = tcpip_io();
-		tcpip_close();
-		tcpip_done();
-		return ret;
+		// start interactive variant, if built-in TCP/IP requested
+		ret = run_tcpip(peers.size(), opt_p, hostname, port);
 	}
 	else if (opt_y != NULL)
 	{
-		// run as replacement for GnuPG et al. (yet-another-openpgp-tool)
-		run_instance(0, time(NULL), opt_e, 0);
-		return ret;
+		// start a single instance as replacement for GnuPG et al.
+		fork_instance(0);
+		ret = wait_instance(0, opt_verbose, pid);
 	}
-
-	// start interactive variant with GNUnet or otherwise a local test
-#ifdef GNUNET
-	static const struct GNUNET_GETOPT_CommandLineOption myoptions[] = {
-		GNUNET_GETOPT_option_flag('C',
-			"clear",
-			"apply cleartext signature framework (cf. RFC 4880)",
-			&gnunet_opt_C
-		),
-		GNUNET_GETOPT_option_uint('e',
-			"expiration",
-			"TIME",
-			"expiration time of generated signature in seconds",
-			&gnunet_opt_sigexptime
-		),
-		GNUNET_GETOPT_option_flag('E',
-			"echo",
-			"enable terminal echo when reading passphrase",
-			&gnunet_opt_E
-		),
-		GNUNET_GETOPT_option_string('H',
-			"hostname",
-			"STRING",
-			"hostname (e.g. onion address) of this peer within PEERS",
-			&gnunet_opt_hostname
-		),
-		GNUNET_GETOPT_option_string('i',
-			"input",
-			"FILENAME",
-			"create signature from FILENAME",
-			&gnunet_opt_ifilename
-		),
-		GNUNET_GETOPT_option_string('k',
-			"keyring",
-			"FILENAME",
-			"use keyring FILENAME containing external revocation keys",
-			&gnunet_opt_k
-		),
-		GNUNET_GETOPT_option_string('o',
-			"output",
-			"FILENAME",
-			"write generated signature to FILENAME",
-			&gnunet_opt_ofilename
-		),
-		GNUNET_GETOPT_option_string('p',
-			"port",
-			"STRING",
-			"GNUnet CADET port to listen/connect",
-			&gnunet_opt_port
-		),
-		GNUNET_GETOPT_option_string('P',
-			"passwords",
-			"STRING",
-			"exchanged passwords to protect private and broadcast channels",
-			&gnunet_opt_passwords
-		),
-		GNUNET_GETOPT_option_string('U',
-			"URI",
-			"STRING",
-			"policy URI tied to signature",
-			&gnunet_opt_URI
-		),
-		GNUNET_GETOPT_option_flag('t',
-			"text",
-			"create canonical text document signature",
-			&gnunet_opt_t
-		),
-		GNUNET_GETOPT_option_flag('V',
-			"verbose",
-			"turn on verbose output",
-			&gnunet_opt_verbose
-		),
-		GNUNET_GETOPT_option_uint('w',
-			"wait",
-			"TIME",
-			"minutes to wait until start of signing protocol",
-			&gnunet_opt_wait
-		),
-		GNUNET_GETOPT_option_uint('W',
-			"aiou-timeout",
-			"TIME",
-			"timeout for point-to-point messages in minutes",
-			&gnunet_opt_W
-		),
-		GNUNET_GETOPT_option_string('y',
-			"yaot",
-			"FILNAME",
-			"yet another OpenPGP tool with private key in FILENAME",
-			&gnunet_opt_y
-		),
-		GNUNET_GETOPT_option_uint('x',
-			"x-tests",
-			NULL,
-			"number of exchange tests",
-			&gnunet_opt_xtests
-		),
-		GNUNET_GETOPT_OPTION_END
-	};
-	ret = GNUNET_PROGRAM_run(argc, argv, usage, about, myoptions, &gnunet_run,
-		argv[0]);
-//	GNUNET_free((void *) argv);
-	if (ret == GNUNET_OK)
-		return 0;
 	else
-		return -1;
+	{
+		// start interactive variant with GNUnet or otherwise a local test
+#ifdef GNUNET
+		static const struct GNUNET_GETOPT_CommandLineOption myoptions[] = {
+			GNUNET_GETOPT_option_flag('C',
+				"clear",
+				"apply cleartext signature framework (cf. RFC 4880)",
+				&gnunet_opt_C
+			),
+			GNUNET_GETOPT_option_uint('e',
+				"expiration",
+				"INTEGER",
+				"expiration time of generated signature in seconds",
+				&gnunet_opt_sigexptime
+			),
+			GNUNET_GETOPT_option_flag('E',
+				"echo",
+				"enable terminal echo when reading passphrase",
+				&gnunet_opt_E
+			),
+			GNUNET_GETOPT_option_string('H',
+				"hostname",
+				"STRING",
+				"hostname (e.g. onion address) of this peer within PEERS",
+				&gnunet_opt_hostname
+			),
+			GNUNET_GETOPT_option_string('i',
+				"input",
+				"FILENAME",
+				"create signature from FILENAME",
+				&gnunet_opt_ifilename
+			),
+			GNUNET_GETOPT_option_string('k',
+				"keyring",
+				"FILENAME",
+				"use keyring FILENAME containing external revocation keys",
+				&gnunet_opt_k
+			),
+			GNUNET_GETOPT_option_string('o',
+				"output",
+				"FILENAME",
+				"write generated signature to FILENAME",
+				&gnunet_opt_ofilename
+			),
+			GNUNET_GETOPT_option_string('p',
+				"port",
+				"STRING",
+				"GNUnet CADET port to listen/connect",
+				&gnunet_opt_port
+			),
+			GNUNET_GETOPT_option_string('P',
+				"passwords",
+				"STRING",
+				"exchanged passwords to protect private and broadcast channels",
+				&gnunet_opt_passwords
+			),
+			GNUNET_GETOPT_option_string('U',
+				"URI",
+				"STRING",
+				"policy URI tied to signature",
+				&gnunet_opt_URI
+			),
+			GNUNET_GETOPT_option_flag('t',
+				"text",
+				"create canonical text document signature",
+				&gnunet_opt_t
+			),
+			GNUNET_GETOPT_option_flag('V',
+				"verbose",
+				"turn on verbose output",
+				&gnunet_opt_verbose
+			),
+			GNUNET_GETOPT_option_uint('w',
+				"wait",
+				"INTEGER",
+				"minutes to wait until start of signing protocol",
+				&gnunet_opt_wait
+			),
+			GNUNET_GETOPT_option_uint('W',
+				"aiou-timeout",
+				"INTEGER",
+				"timeout for point-to-point messages in minutes",
+				&gnunet_opt_W
+			),
+			GNUNET_GETOPT_option_string('y',
+				"yaot",
+				"FILNAME",
+				"yet another OpenPGP tool with private key in FILENAME",
+				&gnunet_opt_y
+			),
+			GNUNET_GETOPT_option_uint('x',
+				"x-tests",
+				NULL,
+				"number of exchange tests",
+				&gnunet_opt_xtests
+			),
+			GNUNET_GETOPT_OPTION_END
+		};
+		ret = GNUNET_PROGRAM_run(argc, argv, usage, about, myoptions,
+			&gnunet_run, argv[0]);
+		if (ret != GNUNET_OK)
+			ret = -1;
 #else
-	std::cerr << "WARNING: GNUnet CADET is required for the message exchange" <<
-		" of this program" << std::endl;
+		ret = run_localtest(peers.size(), opt_verbose, pid, pipefd,
+			broadcast_pipefd, &fork_instance);
 #endif
-
-	std::cerr << "INFO: running local test with " << peers.size() <<
-		" participants" << std::endl;
-	// open pipes
-	for (size_t i = 0; i < peers.size(); i++)
-	{
-		for (size_t j = 0; j < peers.size(); j++)
-		{
-			if (pipe(pipefd[i][j]) < 0)
-				perror("ERROR: dkg-sign (pipe)");
-			if (pipe(broadcast_pipefd[i][j]) < 0)
-				perror("ERROR: dkg-sign (pipe)");
-		}
 	}
-	
-	// start childs
-	for (size_t i = 0; i < peers.size(); i++)
-		fork_instance(i);
-
-	// sleep for five seconds
-	sleep(5);
-	
-	// wait for childs and close pipes
-	for (size_t i = 0; i < peers.size(); i++)
-	{
-		int wstatus = 0;
-		if (opt_verbose)
-			std::cerr << "INFO: waitpid(" << pid[i] << ")" << std::endl;
-		if (waitpid(pid[i], &wstatus, 0) != pid[i])
-			perror("ERROR: dkg-sign (waitpid)");
-		if (!WIFEXITED(wstatus))
-		{
-			std::cerr << "ERROR: protocol instance ";
-			if (WIFSIGNALED(wstatus))
-				std::cerr << pid[i] << " terminated by signal " <<
-					WTERMSIG(wstatus) << std::endl;
-			if (WCOREDUMP(wstatus))
-				std::cerr << pid[i] << " dumped core" << std::endl;
-			ret = -1; // fatal error
-		}
-		else if (WIFEXITED(wstatus))
-		{
-			if (opt_verbose)
-				std::cerr << "INFO: protocol instance " << pid[i] <<
-					" terminated with exit status " << WEXITSTATUS(wstatus) <<
-					std::endl;
-			if (WEXITSTATUS(wstatus))
-				ret = -2; // error
-		}
-		for (size_t j = 0; j < peers.size(); j++)
-		{
-			if ((close(pipefd[i][j][0]) < 0) || (close(pipefd[i][j][1]) < 0))
-				perror("ERROR: dkg-sign (close)");
-			if ((close(broadcast_pipefd[i][j][0]) < 0) ||
-				(close(broadcast_pipefd[i][j][1]) < 0))
-			{
-				perror("ERROR: dkg-sign (close)");
-			}
-		}
-	}
-	
+	if (should_unlock)
+		unlock_memory();
 	return ret;
 }
 
