@@ -1,7 +1,7 @@
 /*******************************************************************************
    This file is part of Distributed Privacy Guard (DKGPG).
 
- Copyright (C) 2018  Heiko Stamer <HeikoStamer@gmx.net>
+ Copyright (C) 2018, 2019  Heiko Stamer <HeikoStamer@gmx.net>
 
    DKGPG is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -47,7 +47,8 @@ int main
 	char		*opt_s = NULL;
 	char		*opt_o = NULL;
 	std::string	sigfrom_str, sigto_str;
-	struct tm	sigfrom_tm = { 0 }, sigto_tm = { 0 };
+	struct tm	sigfrom_tm = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
+	struct tm	sigto_tm = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
 	time_t		sigfrom = 1535000000, sigto = time(NULL);
 
 	// parse command line arguments
@@ -186,7 +187,7 @@ int main
 		filename = arg;
 	}
 #ifdef DKGPG_TESTSUITE
-	filename = "Test1_dkg-pub.asc";
+	filename = "TestTS1_dkg-pub.asc";
 	sfilename = "Test1_output_timestamp.sig";
 	opt_s = (char*)sfilename.c_str();
 	kfilename = "Test1_dkg-pub.asc";
@@ -507,12 +508,12 @@ int main
 		delete ring;
 		return -2;
 	}
-	// 3. key usage flags (signatures made by keys not with the "signing"
+	// 3. key usage flags (signatures made by keys not with the "timestamping"
 	//    capability are not valid)
-	if (!opt_weak && ((keyusage & 0x02) != 0x02))
+	if (!opt_weak && ((keyusage & 0x0800) != 0x0800))
 	{
 		std::cerr << "ERROR: corresponding key of timestamp authority" <<
-			" was not intented for signing" << std::endl;
+			" was not intented for timestamping" << std::endl;
 		delete target_signature;
 		delete signature;
 		delete primary;
