@@ -1,7 +1,7 @@
 /*******************************************************************************
    This file is part of Distributed Privacy Guard (DKGPG).
 
- Copyright (C) 2017, 2018  Heiko Stamer <HeikoStamer@gmx.net>
+ Copyright (C) 2017, 2018, 2019  Heiko Stamer <HeikoStamer@gmx.net>
 
    DKGPG is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -47,7 +47,8 @@ int main
 	char		*opt_k = NULL;
 	char		*opt_s = NULL;
 	std::string	sigfrom_str, sigto_str;
-	struct tm	sigfrom_tm = { 0 }, sigto_tm = { 0 };
+	struct tm	sigfrom_tm = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
+	struct tm	sigto_tm = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 , 0};
 	time_t		sigfrom = 1243810800, sigto = time(NULL);
 
 	// parse command line arguments
@@ -329,8 +330,11 @@ int main
 	TMCG_OpenPGP_Keyring *ring = NULL;
 	if (opt_k)
 	{
+		int opt_verbose_ring = opt_verbose;
+		if (opt_verbose_ring > 0)
+			opt_verbose_ring--;
 		parse_ok = CallasDonnerhackeFinneyShawThayerRFC4880::
-			PublicKeyringParse(armored_pubring, opt_verbose, ring);
+			PublicKeyringParse(armored_pubring, opt_verbose_ring, ring);
 		if (!parse_ok)
 		{
 			std::cerr << "WARNING: cannot use the given keyring" << std::endl;
